"""
SPDX-License-Identifier: MIT

  Copyright (c) 2024, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import json
import os
import subprocess
from dataclasses import dataclass
from typing import Dict, List, Optional, TypedDict

from scanoss.constants import DEFAULT_RETRY, DEFAULT_TIMEOUT
from scanoss.csvoutput import CsvOutput
from scanoss.cyclonedx import CycloneDx
from scanoss.scanossbase import ScanossBase
from scanoss.scanossgrpc import ScanossGrpc
from scanoss.spdxlite import SpdxLite
from scanoss.utils.abstract_presenter import AbstractPresenter

DEFAULT_SYFT_TIMEOUT = 600
DEFAULT_SYFT_COMMAND = 'syft'


@dataclass
class ContainerScannerConfig:
    debug: bool = False
    trace: bool = False
    quiet: bool = False
    retry: int = DEFAULT_RETRY
    timeout: int = DEFAULT_TIMEOUT
    output: Optional[str] = None
    format: Optional[str] = None
    apiurl: Optional[str] = None
    ignore_cert_errors: bool = False
    key: Optional[str] = None
    proxy: Optional[str] = None
    pac: Optional[str] = None
    grpc_proxy: Optional[str] = None
    ca_cert: Optional[str] = None
    syft_command: str = DEFAULT_SYFT_COMMAND
    syft_timeout: int = DEFAULT_SYFT_TIMEOUT
    only_interim_results: bool = False


def create_container_scanner_config_from_args(args) -> ContainerScannerConfig:
    return ContainerScannerConfig(
        debug=args.debug if 'debug' in args else False,
        trace=args.trace if 'trace' in args else False,
        quiet=args.quiet if 'quiet' in args else False,
        retry=args.retry if 'retry' in args else DEFAULT_RETRY,
        timeout=args.timeout if 'timeout' in args else DEFAULT_TIMEOUT,
        output=args.output if 'output' in args else None,
        format=args.format if 'format' in args else None,
        apiurl=args.api2url if 'api2url' in args else None,
        proxy=args.proxy if 'proxy' in args else None,
        pac=args.pac if 'pac' in args else None,
        grpc_proxy=args.grpc_proxy if 'grpc_proxy' in args else None,
        ca_cert=args.ca_cert if 'ca_cert' in args else None,
        ignore_cert_errors=args.ignore_cert_errors if 'ignore_cert_errors' in args else False,
        key=args.key if 'key' in args else None,
        syft_command=args.syft_command if 'syft_command' in args else DEFAULT_SYFT_COMMAND,
        syft_timeout=args.syft_timeout if 'syft_timeout' in args else DEFAULT_SYFT_TIMEOUT,
    )


class LicenseItem(TypedDict):
    value: str
    spdxExpression: str
    type: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


class SyftArtifactItem(TypedDict):
    name: str
    version: str
    type: str
    purl: str
    licenses: List[LicenseItem]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            name=data['name'],
            version=data['version'],
            type=data['type'],
            purl=data['purl'],
            licenses=[LicenseItem.from_dict(lic) for lic in data['licenses']],
        )


class SyftScanResult(TypedDict):
    artifacts: List[SyftArtifactItem]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(artifacts=[SyftArtifactItem.from_dict(a) for a in data['artifacts']])


class PurlItem(TypedDict):
    purl: str
    requirement: Optional[str]


class ContainerScanResultFileItem(TypedDict):
    file: str
    purls: List[PurlItem]


class ContainerScanResult(TypedDict):
    files: List[ContainerScanResultFileItem]


class DependencyLicenseItem(TypedDict):
    value: str
    spdxExpression: str
    type: str


class DependencyItem(TypedDict):
    purl: str
    licenses: List[DependencyLicenseItem]


class DecoratedContainerScanResultFileItem(TypedDict):
    file: str
    id: str
    status: str
    dependencies: List[DependencyItem]


class DecoratedContainerScanResult(TypedDict):
    files: List[ContainerScanResultFileItem]
    status: Dict[str, str]


class SyftScanError(Exception):
    """Base exception for Syft scan errors"""

    pass


class SyftExecutionError(SyftScanError):
    """Raised when Syft returns a non-zero exit code"""

    pass


class SyftJsonError(SyftScanError):
    """Raised when Syft output cannot be parsed as JSON"""

    pass


class SyftTimeoutError(SyftScanError):
    """Raised when a Syft scan times out"""

    pass


class SCANOSSDependencyScanError(Exception):
    """Base exception for SCANOSS dependency scan errors"""

    pass


class DecorateScanResultsError(SCANOSSDependencyScanError):
    """Raised when there is an issue decorating scan results with dependencies"""

    pass


class ContainerScanner:
    """SCANOSS container scanning class.

    This class provides functionality to scan containers using Syft and process
    the results into SCANOSS dependency format.
    """

    def __init__(
        self,
        config: ContainerScannerConfig,
        what_to_scan: str,
    ):
        """Initialize ContainerScanner class.

        Args:
            config: ContainerScannerConfig object containing configuration settings.
        """
        self.base = ScanossBase(
            debug=config.debug,
            trace=config.trace,
            quiet=config.quiet,
        )
        self.presenter = ContainerScannerPresenter(
            self,
            debug=config.debug,
            trace=config.trace,
            quiet=config.quiet,
            output_file=config.output,
            output_format=config.format,
        )
        self.grpc_api = ScanossGrpc(
            debug=config.debug,
            quiet=config.quiet,
            trace=config.trace,
            url=config.apiurl,
            api_key=config.key,
            ca_cert=config.ca_cert,
            proxy=config.proxy,
            pac=config.pac,
            grpc_proxy=config.grpc_proxy,
        )
        self.what_to_scan: str = what_to_scan
        self.syft_command: str = config.syft_command
        self.syft_timeout: int = config.syft_timeout
        self.only_interim_results: bool = config.only_interim_results
        self.syft_output: Optional[SyftScanResult] = None
        self.normalized_syft_output: Optional[ContainerScanResult] = None
        self.decorated_scan_results: Optional[DecoratedContainerScanResult] = None

    def decorate_scan_results_with_dependencies(self) -> None:
        """
        Decorate the scan results with dependencies.
        """
        try:
            decorated_scan_results = self.grpc_api.get_dependencies(self.normalized_syft_output)
            self.decorated_scan_results = decorated_scan_results
            return decorated_scan_results
        except Exception as e:
            error_msg = f'Issue decorating scan results with dependencies: {e}'
            self.base.print_stderr(f'ERROR: {error_msg}')
            raise DecorateScanResultsError(error_msg) from e

    def scan(
        self,
    ) -> ContainerScanResult:
        """Run a syft scan of the specified target.

        Returns:
            ContainerScanResult: The container scan results.

        Raises:
            SyftScanError: For other scan-related errors
        """
        try:
            self.syft_output = self._execute_syft_scan()
            self.normalized_syft_output = self._normalize_syft_output()
            return self.normalized_syft_output
        except Exception as e:
            if isinstance(e, SyftScanError):
                raise
            error_msg = f'Issue running syft scan on {self.what_to_scan}: {e}'
            self.base.print_stderr(f'ERROR: {error_msg}')
            raise SyftScanError(error_msg) from e

    def _execute_syft_scan(self) -> SyftScanResult:
        """
        Execute a Syft scan of the specified target.

        Returns:
            SyftScanResult: The result of the Syft scan.

        Raises:
            SyftScanError: If the Syft scan fails.
            SyftJsonError: If the Syft scan output cannot be parsed as JSON.
            SyftTimeoutError: If the Syft scan times out.
            SyftExecutionError: If the Syft scan execution fails.
        """
        try:
            self.base.print_trace(f'About to execute {self.syft_command} scan {self.what_to_scan} -q -o json')
            self.base.print_msg('Scanning container...')
            result = subprocess.run(
                [self.syft_command, 'scan', self.what_to_scan, '-q', '-o', 'json'],
                cwd=os.getcwd(),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                timeout=self.syft_timeout,
                check=False,
            )
            self.base.print_trace(f'Subprocess return: {result}')

            if result.returncode:
                error_msg = (
                    f'Syft scan of {self.what_to_scan} failed with exit code {result.returncode}:\n{result.stdout}'
                )
                self.base.print_stderr(f'ERROR: {error_msg}')
                raise SyftExecutionError(error_msg)

            try:
                json_data = json.loads(result.stdout)
                return SyftScanResult.from_dict(json_data)
            except json.JSONDecodeError as e:
                error_msg = f'Failed to parse JSON output from syft: {e}\n{result.stdout}'
                self.base.print_stderr(f'ERROR: {error_msg}')
                raise SyftJsonError(error_msg) from e

        except subprocess.TimeoutExpired as e:
            error_msg = f'Timed out attempting to run syft scan on {self.what_to_scan}: {e}'
            self.base.print_stderr(f'ERROR: {error_msg}')
            raise SyftTimeoutError(error_msg) from e

        except Exception as e:
            if isinstance(e, SyftScanError):
                raise
            error_msg = f'Issue running syft scan on {self.what_to_scan}: {e}'
            self.base.print_stderr(f'ERROR: {error_msg}')
            raise SyftScanError(error_msg) from e

    def _get_dependencies(self) -> None:
        """
        Run a dependency scan of the specified target.
        """
        try:
            if not self.normalized_syft_output:
                error_msg = 'Syft scan output is not available'
                self.base.print_stderr(error_msg)
                raise ValueError(error_msg)
            if not self.grpc_api.get_dependencies(self.normalized_syft_output):
                error_msg = 'Failed to get dependencies'
                self.base.print_stderr(error_msg)
                raise SCANOSSDependencyScanError(error_msg)
        except Exception as e:
            error_msg = f'Failed to run dependency scan: {e}'
            self.base.print_stderr(error_msg)
            raise SCANOSSDependencyScanError(error_msg)

    def _normalize_syft_output(self) -> ContainerScanResult:
        """
        Normalize the Syft output data into the same format we use in dependency scanning

        Returns:
            ContainerScanResult: The normalized output
        """
        normalized_output = ContainerScanResult()

        # This is a workaround because we don't have file paths as in dependency scanning, we use the container name
        file_name = self.what_to_scan
        artifacts = self.syft_output['artifacts']

        unique_purls = set()
        unique_purl_items = []

        for artifact in artifacts:
            purl = artifact['purl']
            if purl not in unique_purls:
                unique_purls.add(purl)
                unique_purl_items.append(PurlItem(purl=purl))

        normalized_output['files'] = [
            {
                'file': file_name,
                'purls': unique_purl_items,
            }
        ]

        return normalized_output

    def present(self, output_format: str = None, output_file: str = None):
        """Present the results in the selected format"""
        self.presenter.present(output_format=output_format, output_file=output_file)


class ContainerScannerPresenter(AbstractPresenter):
    """
    ContainerScannerPresenter presenter class
    Handles the presentation of the container scan results
    """

    def __init__(self, scanner: ContainerScanner, **kwargs):
        super().__init__(**kwargs)
        self.scanner = scanner
        self.AVAILABLE_OUTPUT_FORMATS = ['plain', 'cyclonedx', 'spdxlite', 'csv', 'raw']

    def _convert_raw_to_scan_output(self) -> dict:
        """
        Convert the raw output from dependency scanning API to our scan output format

        Returns:
            dict: The converted output
        """
        formatted_output = {}
        if (
            self.scanner.decorated_scan_results
            and 'files' in self.scanner.decorated_scan_results
            and self.scanner.decorated_scan_results['files']
            and isinstance(self.scanner.decorated_scan_results['files'], list)
        ):
            file_item = self.scanner.decorated_scan_results['files'][0]
            if file_item and isinstance(file_item, dict) and 'file' in file_item:
                formatted_output[file_item['file']] = [file_item]

        return formatted_output

    def _format_plain_output(self) -> str:
        """
        Format the scan output data into a plain text string
        """
        return json.dumps(self._convert_raw_to_scan_output(), indent=2)

    def _format_raw_output(self) -> str:
        """
        Format the scan output data into the raw output from dependency scanning API
        """
        if self.scanner.only_interim_results:
            return json.dumps(self.scanner.normalized_syft_output, indent=2)
        return json.dumps(self.scanner.decorated_scan_results, indent=2)

    def _format_cyclonedx_output(self) -> str:
        """
        Format the scan output data into a CycloneDX object
        """
        cdx = CycloneDx(self.base.debug, self.output_file)
        scan_results = {}
        for f in self.scanner.decorated_scan_results['files']:
            scan_results[f['file']] = [f]
        success, cdx_output = cdx.produce_from_json(scan_results)
        if not success:
            error_msg = 'Failed to produce CycloneDX output'
            self.base.print_stderr(error_msg)
            return None
        return json.dumps(cdx_output, indent=2)

    def _format_spdxlite_output(self) -> str:
        """
        Format the scan output data into a SPDXLite object
        """
        spdxlite = SpdxLite(self.base.debug, self.output_file)
        scan_results = {}
        for f in self.scanner.decorated_scan_results['files']:
            scan_results[f['file']] = [f]
        if not spdxlite.produce_from_json(scan_results, self.output_file):
            error_msg = 'Failed to produce SPDXLite output'
            self.base.print_stderr(error_msg)
            raise ValueError(error_msg)

    def _format_csv_output(self) -> str:
        """
        Format the scan output data into a CSV object
        """
        csv = CsvOutput(self.base.debug, self.output_file)
        scan_results = {}
        for f in self.scanner.decorated_scan_results['files']:
            scan_results[f['file']] = [f]
        if not csv.produce_from_json(scan_results, self.output_file):
            error_msg = 'Failed to produce CSV output'
            self.base.print_stderr(error_msg)
            raise ValueError(error_msg)

    def _format_json_output(self) -> str:
        """
        Format the scan output data into a JSON object
        """
        pass



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2025, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import requests

from ..scanossbase import ScanossBase

HTTP_OK = 200

class DependencyTrackService(ScanossBase):

    def __init__(
            self,
            api_key: str,
            url: str,
            debug: bool = False,
            trace: bool = False,
            quiet: bool = False,
    ):
        super().__init__(debug=debug, trace=trace, quiet=quiet)
        if not url:
            raise ValueError("Error: Dependency Track URL is required")
        self.url = url.strip().rstrip('/')
        if not api_key:
            raise ValueError("Error: Dependency Track API key is required")
        self.api_key = api_key

    def get_project_by_name_version(self, name, version):
        """
        Get project information by name and version from Dependency Track

        Args:
            name: Project name to search for
            version: Project version to search for

        Returns:
            dict: Project data if found, None otherwise
        """
        if not name or not version:
            self.print_stderr('Error: Missing name or version.')
            return None
        # Use the project search endpoint
        params = {
            'name': name,
            'version': version
        }
        self.print_debug(f'Searching for project by: {params}')
        return self.get_dep_track_data(f'{self.url}/api/v1/project/lookup', params)

    def get_project_status(self, upload_token):
        """
        Get Dependency Track project processing status.

        Queries the Dependency Track API to check if the project upload
        processing is complete using the upload token.

        Returns:
            dict: Project status information or None if request fails
        """
        if not upload_token:
            self.print_stderr('Error: Missing upload token. Cannot search for project status.')
            return None
        self.print_trace(f'URL: {self.url} Upload token: {upload_token}')
        return self.get_dep_track_data(f'{self.url}/api/v1/event/token/{upload_token}')

    def get_project_violations(self,project_id:str):
        """
        Get project violations from Dependency Track.

        Waits for project processing to complete, then retrieves all policy
        violations for the specified project ID.

        Returns:
        List of policy violations or None if the request fails
        """
        if not project_id:
            self.print_stderr('Error: Missing project id. Cannot search for project violations.')
            return None
        # Return the result as-is - None indicates API failure, empty list means no violations
        return self.get_dep_track_data(f'{self.url}/api/v1/violation/project/{project_id}')

    def get_project_by_id(self, project_id:str):
        """
        Get a Dependency Track project by id.

        Queries the Dependency Track API to get a project by id

        Returns:
            dict
        """
        if not project_id:
            self.print_stderr('Error: Missing project id. Cannot search for project.')
            return None
        self.print_trace(f'URL: {self.url}, UUID: {project_id}')
        return self.get_dep_track_data(f'{self.url}/api/v1/project/{project_id}')

    def get_dep_track_data(self, uri, params=None):
        if not uri:
            self.print_stderr('Error: Missing URI. Cannot search for project.')
            return None
        req_headers = {'X-Api-Key': self.api_key, 'Content-Type': 'application/json'}
        try:
            if params:
                response = requests.get(uri, headers=req_headers, params=params)
            else:
                response = requests.get(uri, headers=req_headers)
            response.raise_for_status()  # Raises an HTTPError for bad responses
            return response.json()
        except requests.exceptions.RequestException as e:
            self.print_stderr(f"Error: Problem getting project data: {e}")
        return None



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2025, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import base64
import json
import traceback

import requests

from ..cyclonedx import CycloneDx
from ..scanossbase import ScanossBase
from ..services.dependency_track_service import DependencyTrackService
from ..utils.file import validate_json_file


def _build_payload(encoded_sbom: str, project_id, project_name, project_version) -> dict:
    """
    Build the API payload

    Args:
        encoded_sbom: Base64 encoded SBOM

    Returns:
        API payload dictionary
    """
    if project_id:
        return {'project': project_id, 'bom': encoded_sbom}
    else:
        return {
            'projectName': project_name,
            'projectVersion': project_version,
            'autoCreate': True,
            'bom': encoded_sbom,
        }


class DependencyTrackExporter(ScanossBase):
    """
    Class for exporting SBOM files to Dependency Track
    """
    def __init__( # noqa: PLR0913
        self,
        url: str = None,
        apikey: str = None,
        output: str = None,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False
    ):
        """
        Initialize DependencyTrackExporter

        Args:
            url: Dependency Track URL
            apikey: Dependency Track API Key
            output: File to store output response data (optional)
            debug: Enable debug output
            trace: Enable trace output
            quiet: Enable quiet mode
        """
        super().__init__(debug=debug, trace=trace, quiet=quiet)
        self.url = url.rstrip('/')
        self.apikey = apikey
        self.output = output
        self.dt_service = DependencyTrackService(self.apikey, self.url, debug=debug, trace=trace, quiet=quiet)

    def _read_and_validate_sbom(self, input_file: str) -> dict:
        """
        Read and validate the SBOM file

        Args:
            input_file: Path to the SBOM file

        Returns:
            Parsed SBOM content as dictionary

        Raises:
            ValueError: If the file doesn't exist or is invalid or not a valid CycloneDX SBOM
        """
        result = validate_json_file(input_file)
        if not result.is_valid:
            raise ValueError(f'Invalid JSON file: {result.error}')

        cdx = CycloneDx(debug=self.debug)
        if not cdx.is_cyclonedx_json(json.dumps(result.data)):
            raise ValueError(f'Input file is not a valid CycloneDX SBOM: {input_file}')
        return result.data

    def _encode_sbom(self, sbom_content: dict) -> str:
        """
        Encode SBOM content to base64

        Args:
            sbom_content: SBOM dictionary

        Returns:
            Base64 encoded string
        """
        if not sbom_content:
            self.print_stderr('Warning: Empty SBOM content provided')
            return ''
        # Check if SBOM has no components (empty scan results)
        components = sbom_content.get('components', [])
        if len(components) == 0:
            self.print_msg('Notice: SBOM contains no components (empty scan results)')
        json_str = json.dumps(sbom_content, separators=(',', ':'))
        encoded = base64.b64encode(json_str.encode('utf-8')).decode('utf-8')
        return encoded

    def upload_sbom_file(self, input_file, project_id, project_name, project_version, output_file):
        """
        Uploads an SBOM file to the specified project with an
        optional output file and processes the file content for validation.

        Args:
            input_file (str): The path to the SBOM file to be read and uploaded.
            project_id (str): The unique identifier of the project to which the SBOM is being uploaded.
            project_name (str): The name of the project to which the SBOM is being uploaded.
            project_version (str): The version of the project to which the SBOM is being uploaded.
            output_file (str): The path to save output related to the SBOM upload process.

        Returns:
            bool: Returns True if the SBOM file was uploaded successfully, False otherwise.

        Raises:
            ValueError: Raised if there are validation issues with the SBOM content.
        """
        try:
            if not self.quiet:
                self.print_stderr(f'Reading SBOM file: {input_file}')
            sbom_content = self._read_and_validate_sbom(input_file)
            return self.upload_sbom_contents(sbom_content, project_id, project_name, project_version, output_file)
        except ValueError as e:
            self.print_stderr(f'Validation error: {e}')
        return False

    def upload_sbom_contents(self, sbom_content: dict, project_id, project_name, project_version, output_file) -> bool:
        """
        Uploads an SBOM to a Dependency Track server.

        Parameters:
            sbom_content (dict): The SBOM content in dictionary format to be uploaded.
            project_id: The unique identifier for the project.
            project_name: The name of the project in Dependency Track.
            project_version: The version of the project in Dependency Track.
            output_file: The path to the file where the token and UUID data
                should be written. If not provided, the data will be written to
                standard output.

        Returns:
            bool: True if the upload is successful; False otherwise.

        Raises:
            ValueError: If the SBOM encoding process fails.
            requests.exceptions.RequestException: If an error occurs during the HTTP request.
            Exception: For any other unexpected error.
        """
        if not project_id and not (project_name and project_version):
            self.print_stderr('Error: Missing project id or name and version.')
            return False
        output = self.output
        if output_file:
            output = output_file
        try:
            self.print_debug('Encoding SBOM to base64')
            payload = _build_payload(self._encode_sbom(sbom_content), project_id, project_name, project_version)
            url = f'{self.url}/api/v1/bom'
            headers = {'Content-Type': 'application/json', 'X-Api-Key': self.apikey}
            self.print_trace(f'URL: {url}, Headers: {headers}, Payload keys: {list(payload.keys())}')
            self.print_msg('Uploading SBOM to Dependency Track...')
            response = requests.put(url, json=payload, headers=headers)
            response.raise_for_status()
            # Treat any 2xx status as success
            if (requests.codes.ok <= response.status_code < requests.codes.multiple_choices and
                    response.status_code != requests.codes.no_content):
                self.print_msg('SBOM uploaded successfully')
                try:
                    response_data = response.json()
                    token = ''
                    project_uuid = project_id
                    if 'token' in response_data:
                        token = response_data['token']
                    if project_name and project_version:
                      project_data = self.dt_service.get_project_by_name_version(project_name, project_version)
                      if project_data:
                         project_uuid = project_data.get("uuid", project_id)
                    token_json = json.dumps(
                        {"token": token, "project_uuid": project_uuid},
                        indent=2
                    )
                    self.print_to_file_or_stdout(token_json, output)
                except json.JSONDecodeError:
                    pass
                return True
            else:
                self.print_stderr(f'Upload failed with status code: {response.status_code}')
                self.print_stderr(f'Response: {response.text}')
        except ValueError as e:
            self.print_stderr(f'DT SBOM Upload Validation error: {e}')
        except requests.exceptions.RequestException as e:
            self.print_stderr(f'DT API Request error: {e}')
        except Exception as e:
            self.print_stderr(f'Unexpected error: {e}')
            if self.debug:
                traceback.print_exc()
        return False



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2021, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import datetime
import json
import os.path
import sys
import uuid

from cyclonedx.schema import SchemaVersion
from cyclonedx.validation.json import JsonValidator

from . import __version__
from .scanossbase import ScanossBase
from .spdxlite import SpdxLite


class CycloneDx(ScanossBase):
    """
    CycloneDX management class
    Handle all interaction with CycloneDX formatting
    """

    def __init__(self, debug: bool = False, output_file: str = None):
        """
        Initialise the CycloneDX class
        """
        super().__init__(debug)
        self.output_file = output_file
        self.debug = debug
        self._spdx = SpdxLite(debug=debug)

    def parse(self, data: dict):  # noqa: PLR0912, PLR0915
        """
        Parse the given input (raw/plain) JSON string and return CycloneDX summary
        :param data: dict - JSON object
        :return: CycloneDX dictionary, and vulnerability dictionary
        """
        if data is None:
            self.print_stderr('ERROR: No JSON data provided to parse.')
            return None, None
        if len(data) == 0:
            self.print_msg('Warning: Empty scan results provided. Returning empty component dictionary.')
            return {}, {}
        self.print_debug('Processing raw results into CycloneDX format...')
        cdx = {}
        vdx = {}
        for f in data:
            file_details = data.get(f)
            # print(f'File: {f}: {file_details}')
            for d in file_details:
                id_details = d.get('id')
                if not id_details or id_details == 'none':
                    continue
                purl = None
                if id_details == 'dependency':
                    dependencies = d.get('dependencies')
                    if not dependencies:
                        self.print_stderr(f'Warning: No Dependencies found for {f}: {file_details}')
                        continue
                    for deps in dependencies:
                        purl = deps.get('purl')
                        if not purl:
                            self.print_stderr(f'Warning: No PURL found for {f}: {deps}')
                            continue
                        if cdx.get(purl):
                            self.print_debug(f'Component {purl} already stored: {cdx.get(purl)}')
                            continue
                        fd = {}
                        for field in ['component', 'version']:
                            fd[field] = deps.get(field, '')
                        licenses = deps.get('licenses')
                        fdl = []
                        if licenses:
                            dc = []
                            for lic in licenses:
                                name = lic.get('name')
                                if name not in dc:  # Only save the license name once
                                    fdl.append({'id': name})
                                    dc.append(name)
                        fd['licenses'] = fdl
                        cdx[purl] = fd
                else:
                    purls = d.get('purl')
                    if not purls:
                        self.print_stderr(f'Purl block missing for {f}: {file_details}')
                        continue
                    for p in purls:
                        self.print_debug(f'Purl: {p}')
                        purl = p
                        break
                    if not purl:
                        self.print_stderr(f'Warning: No PURL found for {f}: {file_details}')
                        continue
                    fd = {}
                    vulnerabilities = d.get('vulnerabilities')
                    if vulnerabilities:
                        for vuln in vulnerabilities:
                            vuln_id = vuln.get('ID')
                            if vuln_id == '':
                                vuln_id = vuln.get('id')
                            if not vuln_id or vuln_id == '':  # Skip empty ids
                                continue
                            vuln_cve = vuln.get('CVE', '')
                            if vuln_cve == '':
                                vuln_cve = vuln.get('cve', '')
                            if vuln_id.upper().startswith('CPE:'):
                                fd['cpe'] = vuln_id  # Save the component CPE if we have one
                                if vuln_cve != '':
                                    vuln_id = vuln_cve
                            vd = vdx.get(vuln_id)  # Check if we've already encountered this vulnerability
                            if not vd:
                                vuln_source = vuln.get('source', '').lower()
                                vd = {
                                    'cve': vuln_cve,
                                    'source': 'NVD' if vuln_source == 'nvd' else 'GitHub Advisories',
                                    'url': f'https://nvd.nist.gov/vuln/detail/{vuln_cve}'
                                    if vuln_source == 'nvd'
                                    else f'https://github.com/advisories/{vuln_id}',
                                    'severity': self._sev_lookup(vuln.get('severity', 'unknown').lower()),
                                    'affects': set(),
                                }
                            vd.get('affects').add(purl)
                            vdx[vuln_id] = vd
                    if cdx.get(purl):
                        self.print_debug(f'Component {purl} already stored: {cdx.get(purl)}')
                        continue
                    for field in ['id', 'vendor', 'component', 'version', 'latest']:
                        fd[field] = d.get(field)
                    licenses = d.get('licenses')
                    fdl = []
                    if licenses:
                        for lic in licenses:
                            fdl.append({'id': lic.get('name')})
                    fd['licenses'] = fdl
                    cdx[purl] = fd
        # self.print_stderr(f'VD: {vdx}')
        # self.print_stderr(f'CDX: {cdx}')
        return cdx, vdx

    def produce_from_file(self, json_file: str, output_file: str = None) -> bool:
        """
        Parse plain/raw input JSON file and produce CycloneDX output
        :param json_file:
        :param output_file:
        :return: True if successful, False otherwise
        """
        if not json_file:
            self.print_stderr('ERROR: No JSON file provided to parse.')
            return False
        if not os.path.isfile(json_file):
            self.print_stderr(f'ERROR: JSON file does not exist or is not a file: {json_file}')
            return False
        with open(json_file, 'r') as f:
            success = self.produce_from_str(f.read(), output_file)
        return success

    def produce_from_json(self, data: dict, output_file: str = None) -> tuple[bool, dict]:  # noqa: PLR0912
        """
        Produce the CycloneDX output from the raw scan results input data

        Args:
            data (dict): JSON object
            output_file (str, optional): Output file (optional). Defaults to None.

        Returns:
            bool: True if successful, False otherwise
            json: The CycloneDX output
        """
        cdx, vdx = self.parse(data)
        if cdx is None:
            self.print_stderr('ERROR: No CycloneDX data returned for the JSON string provided.')
            return False, {}
        if len(cdx) == 0:
            self.print_msg('Warning: Empty scan results - generating minimal CycloneDX SBOM with no components.')
        self._spdx.load_license_data()  # Load SPDX license name data for later reference
        #
        # Using CDX version 1.4: https://cyclonedx.org/docs/1.4/json/
        # Validate using: https://github.com/CycloneDX/cyclonedx-cli
        # cyclonedx-cli validate --input-format json --input-version v1_4 --fail-on-errors --input-file cdx.json
        #
        data = {
            'bomFormat': 'CycloneDX',
            'specVersion': '1.4',
            'serialNumber': f'urn:uuid:{uuid.uuid4()}',
            'version': 1,
            'metadata': {
                'timestamp': datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%dT%H:%M:%SZ'),
                'tools': [
                    {
                        'vendor': 'SCANOSS',
                        'name': 'scanoss-py',
                        'version': __version__,
                    }
                ],
                'component': {'type': 'application', 'name': 'NOASSERTION', 'version': 'NOASSERTION'},
            },
            'components': [],
            'vulnerabilities': [],
        }
        for purl in cdx:
            comp = cdx.get(purl)
            lic_text = []
            licenses = comp.get('licenses')
            if licenses:
                lic_set = set()
                for lic in licenses:  # Get a unique set of licenses
                    lc_id = lic.get('id')
                    if not lc_id:
                        continue
                    spdx_id = self._spdx.get_spdx_license_id(lc_id)
                    lic_set.add(spdx_id if spdx_id else lc_id)
                for lc_id in lic_set:  # Store licenses for later inclusion
                    spdx_id = self._spdx.get_spdx_license_id(lc_id)
                    if not spdx_id:
                        lic_text.append({'license': {'name': lc_id}})  # Not an SPDX license, so store it by name
                    else:
                        lic_text.append({'license': {'id': spdx_id}})
            c_data = {
                'type': 'library',
                'name': comp.get('component'),
                'publisher': comp.get('vendor', ''),
                'version': comp.get('version'),
                'purl': purl,
                'bom-ref': purl,
                'licenses': lic_text,
            }
            cpe = comp.get('cpe', '')
            if cpe and cpe != '':
                c_data['cpe'] = cpe
            data['components'].append(c_data)
        # End for loop
        if vdx:
            for vuln_id in vdx:
                vulns = vdx.get(vuln_id)
                if not vulns:
                    continue
                v_source = vulns.get('source')
                affects = []
                for purl in vulns.get('affects'):
                    affects.append({'ref': purl})
                vd = {
                    'id': vuln_id,
                    'source': {'name': v_source, 'url': vulns.get('url')},
                    'ratings': [{'severity': vulns.get('severity', 'unknown')}],
                    'affects': affects,
                }
                data['vulnerabilities'].append(vd)
            # End for loop

        file = sys.stdout
        if not output_file and self.output_file:
            output_file = self.output_file
        if output_file:
            file = open(output_file, 'w')
        print(json.dumps(data, indent=2), file=file)
        if output_file:
            file.close()

        return True, data

    def produce_from_str(self, json_str: str, output_file: str = None) -> bool:
        """
        Produce CycloneDX output from input JSON string
        :param json_str: input JSON string
        :param output_file: Output file (optional)
        :return: True if successful, False otherwise
        """
        if not json_str:
            self.print_stderr('ERROR: No JSON string provided to parse.')
            return False
        try:
            data = json.loads(json_str)
        except Exception as e:
            self.print_stderr(f'ERROR: Problem parsing input JSON: {e}')
            return False
        return self.produce_from_json(data, output_file)

    def _normalize_vulnerability_id(self, vuln: dict) -> tuple[str, str]:
        """
        Normalize vulnerability ID and CVE from different possible field names.
        Returns tuple of (vuln_id, vuln_cve).
        """
        vuln_id = vuln.get('ID', '') or vuln.get('id', '')
        vuln_cve = vuln.get('CVE', '') or vuln.get('cve', '')

        # Skip CPE entries, use CVE if available
        if vuln_id.upper().startswith('CPE:') and vuln_cve:
            vuln_id = vuln_cve

        return vuln_id, vuln_cve

    def _create_vulnerability_entry(self, vuln_id: str, vuln: dict, vuln_cve: str, purl: str) -> dict:
        """
        Create a new vulnerability entry for CycloneDX format.
        """
        vuln_source = vuln.get('source', '').lower()
        return {
            'id': vuln_id,
            'source': {
                'name': 'NVD' if vuln_source == 'nvd' else 'GitHub Advisories',
                'url': f'https://nvd.nist.gov/vuln/detail/{vuln_cve}'
                if vuln_source == 'nvd'
                else f'https://github.com/advisories/{vuln_id}',
            },
            'ratings': [{'severity': self._sev_lookup(vuln.get('severity', 'unknown').lower())}],
            'affects': [{'ref': purl}],
        }

    def append_vulnerabilities(self, cdx_dict: dict, vulnerabilities_data: dict, purl: str) -> dict:
        """
        Append vulnerabilities to an existing CycloneDX dictionary

        Args:
            cdx_dict (dict): The existing CycloneDX dictionary
            vulnerabilities_data (dict): The vulnerabilities data from get_vulnerabilities_json
            purl (str): The PURL of the component these vulnerabilities affect

        Returns:
            dict: The updated CycloneDX dictionary with vulnerabilities appended
        """
        if not cdx_dict or not vulnerabilities_data:
            return cdx_dict

        if 'vulnerabilities' not in cdx_dict:
            cdx_dict['vulnerabilities'] = []

        # Extract vulnerabilities from the response
        vulns_list = vulnerabilities_data.get('purls', [])
        if not vulns_list:
            return cdx_dict

        vuln_items = vulns_list[0].get('vulnerabilities', [])

        for vuln in vuln_items:
            vuln_id, vuln_cve = self._normalize_vulnerability_id(vuln)

            # Skip empty IDs or CPE-only entries
            if not vuln_id or vuln_id.upper().startswith('CPE:'):
                continue

            # Check if vulnerability already exists
            existing_vuln = next((v for v in cdx_dict['vulnerabilities'] if v.get('id') == vuln_id), None)

            if existing_vuln:
                # Add this PURL to the affects list if not already present
                if not any(ref.get('ref') == purl for ref in existing_vuln.get('affects', [])):
                    existing_vuln['affects'].append({'ref': purl})
            else:
                # Create new vulnerability entry
                cdx_dict['vulnerabilities'].append(self._create_vulnerability_entry(vuln_id, vuln, vuln_cve, purl))

        return cdx_dict

    @staticmethod
    def _sev_lookup(value: str):
        """
        Lookup the given severity and return a CycloneDX valid version
        :param value: severity to lookup
        :return: CycloneDX severity
        """
        return {
            'critical': 'critical',
            'high': 'high',
            'medium': 'medium',
            'moderate': 'medium',
            'low': 'low',
            'info': 'info',
            'none': 'none',
            'unknown': 'unknown',
        }.get(value, 'unknown')

    def is_cyclonedx_json(self, json_string: str) -> bool:
        """
        Validate if the given JSON string is a valid CycloneDX JSON string
        Args:
            json_string (str): JSON string to validate
        Returns:
            bool: True if the JSON string is valid, False otherwise
        """
        try:
            cdx_json_validator = JsonValidator(SchemaVersion.V1_6)
            json_validation_errors = cdx_json_validator.validate_str(json_string)
            if json_validation_errors:
                self.print_stderr(f'ERROR: Problem parsing input JSON: {json_validation_errors}')
                return False
            return True
        except Exception as e:
            self.print_stderr(f'ERROR: Problem parsing input JSON: {e}')
            return False


#
# End of CycloneDX Class
#



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2025, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import struct
from typing import List


class CRC64:
    """
    CRC64 ECMA implementation matching Go's hash/crc64 package.
    Uses polynomial: 0xC96C5795D7870F42
    """

    POLY = 0xC96C5795D7870F42
    _TABLE = None

    def __init__(self):
        if CRC64._TABLE is None:
            CRC64._TABLE = self._make_table()
        self.crc = 0xFFFFFFFFFFFFFFFF  # Initial value

    def _make_table(self) -> list:
        """Generate the CRC64 lookup table."""
        table = []
        for i in range(256):
            crc = i
            for _ in range(8):
                if crc & 1:
                    crc = (crc >> 1) ^ self.POLY
                else:
                    crc >>= 1
            table.append(crc)
        return table

    def update(self, data: bytes) -> None:
        """Update the CRC with new data."""
        if isinstance(data, str):
            data = data.encode('utf-8')

        crc = self.crc
        for b in data:
            crc = (crc >> 8) ^ CRC64._TABLE[(crc ^ b) & 0xFF]  # Use class-level table
        self.crc = crc

    def digest(self) -> int:
        """Get the current CRC value."""
        return self.crc ^ 0xFFFFFFFFFFFFFFFF  # Final XOR value

    def hexdigest(self):
        """Get the current CRC value as a hexadecimal string."""
        return format(self.digest(), '016x')

    @classmethod
    def checksum(cls, data: bytes) -> int:
        """Calculate CRC64 checksum for the given data."""
        crc = cls()
        crc.update(data)
        return crc.digest()

    @classmethod
    def get_hash_buff(cls, buff: bytes) -> List[bytes]:
        """
        Get the hash value of the given buffer, and converts it to 8 bytes in big-endian order.

        Args:
            buff (bytes): The buffer to get the hash value of.

        Returns:
            bytes: The hash value of the given buffer, and converts it to 8 bytes in big-endian order.
        """
        crc = cls()
        crc.update(buff)
        hash_val = crc.digest()

        return list(struct.pack('>Q', hash_val))



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2025, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import json
import os.path
from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, TypeVar

from ..policy_check import PolicyCheck
from ..utils.license_utils import LicenseUtil


class ComponentID(Enum):
    """
    Enumeration representing different types of software components.

    Attributes:
        FILE (str): Represents a file component (value: "file").
        SNIPPET (str): Represents a code snippet component (value: "snippet").
        DEPENDENCY (str): Represents a dependency component (value: "dependency").
    """

    FILE = 'file'
    SNIPPET = 'snippet'
    DEPENDENCY = 'dependency'


#
# End of ComponentID Class
#

T = TypeVar('T')
class RawBase(PolicyCheck[T]):
    """
    A base class to perform inspections over scan results.

    This class provides a basic for scan results inspection, including methods for
    processing scan results components and licenses.

    Inherits from:
        ScanossBase: A base class providing common functionality for SCANOSS-related operations.
    """

    def __init__( # noqa: PLR0913
        self,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
        format_type: str = None,
        filepath: str = None,
        output: str = None,
        status: str = None,
        name: str = None,
    ):
        super().__init__(debug, trace, quiet, format_type,status, name, output)
        self.license_util = LicenseUtil()
        self.filepath = filepath
        self.output = output
        self.results = self._load_input_file()

    @abstractmethod
    def _get_components(self):
        """
        Retrieve and process components from the preloaded results.

        This method performs the following steps:
        1. Checks if the results have been previously loaded (self.results).
        2. Extracts and processes components from the loaded results.

        :return: A list of processed components, or None if an error occurred during any step.

        Possible reasons for returning None include:
        - Results not loaded (self.results is None)
        - Failure to extract components from the results

        Note:
        - This method assumes that the results have been previously loaded and stored in self.results.
        - Implementations must extract components (e.g. via `_get_components_data`,
          `_get_dependencies_data`, or other helpers).
        - If `self.results` is `None`, simply return `None`.
        """
    pass

    def _append_component(self, components: Dict[str, Any], new_component: Dict[str, Any]) -> Dict[str, Any]:
        """
          Append a new component to the component dictionary.

          This function creates a new entry in the component dictionary for the given component,
          initializing all required counters:
          - count: Total occurrences of this component (used by both license and component summaries)
          - declared: Number of times this component is marked as 'identified' (used by component summary)
          - undeclared: Number of times this component is marked as 'pending' (used by component summary)

          Each component also contains a 'licenses' dictionary where each license entry tracks:
          - count: Number of times this license appears for this component (used by license summary)

          Args:
              components: The existing dictionary of components
              new_component: The new component to be added
          Returns:
              The updated components dictionary
          """
        match_id = new_component.get('id')
        # Determine the component key and purl based on component type
        if match_id in [ComponentID.FILE.value, ComponentID.SNIPPET.value]:
            purl = new_component['purl'][0]  # Take the first purl for these component types
        else:
            purl = new_component['purl']

        if not purl:
            self.print_debug(f'WARNING: _append_component: No purl found for new component: {new_component}')
            return components

        component_key = f'{purl}@{new_component["version"]}'
        status = new_component.get('status')

        if component_key in components:
            # Component already exists, update component counters and try to append a new license
            self._update_component_counters(components[component_key], status)
            self._append_license_to_component(components, new_component, component_key)
            # Maintain 'pending' status - takes precedence over 'identified'
            if status == 'pending':
                components[component_key]['status'] = "pending"
            return components

        # Create a new component
        components[component_key] = {
            'purl': purl,
            'version': new_component['version'],
            'licenses': {},
            'status': status,
            'count': 1,
            'declared': 1 if status == 'identified' else 0,
            'undeclared': 1 if status == 'pending' else 0
        }

        ## Append license to component
        self._append_license_to_component(components, new_component, component_key)
        return components

    def _append_license_to_component(self,
         components: Dict[str, Any], new_component: Dict[str, Any], component_key: str) -> None:
        """
        Add or update licenses for an existing component.

        For each license in the component:
        - If the license already exists, increments its count
        - If it's a new license, adds it with an initial count of 1

        The license count is used by license_summary to track how many times each license appears
        across all components. This count contributes to:
        - Total number of licenses in the project
        - Number of copyleft licenses when the license is marked as copyleft

        Args:
            components: Dictionary containing all components
            new_component: Component whose licenses need to be processed
            component_key: purl + version of the component to be updated
        """
        # If not licenses are present
        if not new_component.get('licenses'):
            self.print_debug(f'WARNING: Results missing licenses. Skipping: {new_component}')
            return

        licenses_order_by_source_priority = self._get_licenses_order_by_source_priority(new_component['licenses'])
        # Process licenses for this component
        for license_item in licenses_order_by_source_priority:
            if license_item.get('name'):
                spdxid = license_item['name']
                source = license_item.get('source')
                if not source:
                    source = 'unknown'

                if spdxid in components[component_key]['licenses']:
                    # If license exists, increment counter
                    components[component_key]['licenses'][spdxid]['count'] += 1 # Increment counter for license
                else:
                    # If a license doesn't exist, create new entry
                    components[component_key]['licenses'][spdxid] = {
                        'spdxid': spdxid,
                        'copyleft': self.license_util.is_copyleft(spdxid),
                        'url': self.license_util.get_spdx_url(spdxid),
                        'source': source,
                        'count': 1, # Set counter to 1 on new license
                    }

    def _update_component_counters(self, component, status):
        """Update component counters based on status."""
        component['count'] += 1
        if status == 'identified':
            component['declared'] += 1
        else:
            component['undeclared'] += 1

    def _get_components_data(self, results: Dict[str, Any], components: Dict[str, Any]) -> Dict[str, Any]:
        """
           Extract and process file and snippet components from results.

           This method processes scan results to build or update component entries. For each component:

           Component Counters (used by ComponentSummary):
           - count: Incremented for each occurrence of the component
           - declared: Incremented when component status is 'identified'
           - undeclared: Incremented when component status is 'pending'

           License Tracking:
           - For new components, initializes license dictionary through _append_component
           - For existing components, updates license counters through _append_license_to_component
             which tracks the number of occurrences of each license

           Args:
               results: A dictionary containing the raw results of a component scan
           Returns:
               Updated components dictionary with file and snippet data
           """
        for component in results.values():
            for c in component:
                component_id = c.get('id')
                if not component_id:
                    self.print_debug(f'WARNING: Result missing id. Skipping: {c}')
                    continue
                ## Skip dependency
                if component_id == ComponentID.DEPENDENCY.value:
                    continue
                status = c.get('status')
                if not status:
                    self.print_debug(f'WARNING: Result missing status. Skipping: {c}')
                    continue
                if component_id in [ComponentID.FILE.value, ComponentID.SNIPPET.value]:
                    if not c.get('purl'):
                        self.print_debug(f'WARNING: Result missing purl. Skipping: {c}')
                        continue
                    if len(c.get('purl')) <= 0:
                        self.print_debug(f'WARNING: Result missing purls. Skipping: {c}')
                        continue
                    version = c.get('version')
                    if not version:
                        self.print_debug(f'WARNING: Result missing version. Setting it to unknown: {c}')
                        version = 'unknown'
                        c['version'] = version #If no version exists. Set 'unknown' version to current component
                    # Append component
                    components = self._append_component(components, c)

            # End component loop
        # End components loop
        return components

    def _get_dependencies_data(self, results: Dict[str, Any], components: Dict[str, Any]) -> Dict[str, Any]:
        """
        Extract and process dependency components from results.

        :param results: A dictionary containing the raw results of a component scan
        :param components: Existing components dictionary to update
        :return: Updated components dictionary with dependency data
        """
        for component in results.values():
            for c in component:
                component_id = c.get('id')
                if not component_id:
                    self.print_debug(f'WARNING: Result missing id. Skipping: {c}')
                    continue
                status = c.get('status')
                if not status:
                    self.print_debug(f'WARNING: Result missing status. Skipping: {c}')
                    continue
                if component_id == ComponentID.DEPENDENCY.value:
                    if c.get('dependencies') is None:
                        continue
                    for dependency in c['dependencies']:
                        if not dependency.get('purl'):
                            self.print_debug(f'WARNING: Dependency result missing purl. Skipping: {dependency}')
                            continue
                        version = dependency.get('version')
                        if not version:
                            self.print_debug(f'WARNING: Result missing version. Setting it to unknown: {c}')
                            version = 'unknown'
                            c['version'] = version  # Set an 'unknown' version to the current component

                        # Append component
                        components = self._append_component(components, dependency)

                    # End dependency loop
            # End component loop
        # End of result loop
        return components

    def _load_input_file(self):
        """
        Load the result.json file

          Returns:
              Dict[str, Any]: The parsed JSON data
        """
        if not os.path.exists(self.filepath):
            self.print_stderr(f'ERROR: The file "{self.filepath}" does not exist.')
            return None
        with open(self.filepath, 'r') as jsonfile:
            try:
                return json.load(jsonfile)
            except Exception as e:
                self.print_stderr(f'ERROR: Problem parsing input JSON: {e}')
        return None

    def _convert_components_to_list(self, components: dict):
        if components is None:
            self.print_debug(f'WARNING: Components is empty {self.results}')
            return None
        results_list = list(components.values())
        for component in results_list:
            licenses = component.get('licenses')
            if licenses is not None:
                component['licenses'] = list(licenses.values())
            else:
                self.print_debug(f'WARNING: Licenses missing for: {component}')
                component['licenses'] = []
        return results_list

    def _get_licenses_order_by_source_priority(self,licenses_data):
        """
        Select licenses based on source priority:
        1. component_declared (highest priority)
        2. license_file
        3. file_header
        4. scancode (lowest priority)

        If any high-priority source is found, return only licenses from that source.
        If none found, return all licenses.

        Returns: list with ordered licenses by source.
        """
        # Define priority order (highest to lowest)
        priority_sources = ['component_declared', 'license_file', 'file_header', 'scancode']

        # Group licenses by source
        licenses_by_source = {}
        for license_item in licenses_data:

            source = license_item.get('source', 'unknown')
            if source not in licenses_by_source:
                licenses_by_source[source] = {}

            license_name = license_item.get('name')
            if license_name:
                # Use license name as key, store full license object as value
                # If duplicate license names exist in same source, the last one wins
                licenses_by_source[source][license_name] = license_item

        # Find the highest priority source that has licenses
        for priority_source in priority_sources:
            if priority_source in licenses_by_source:
                self.print_trace(f'Choosing {priority_source} as source')
                return list(licenses_by_source[priority_source].values())

        # If no priority sources found, combine all licenses into a single list
        self.print_debug("No priority sources found, returning all licenses as list")
        return licenses_data

    def _group_components_by_license(self,components):
        """
        Groups components by their unique component-license pairs.

        This method processes a list of components and creates unique entries for each
        component-license combination. If a component has multiple licenses, it will create
        separate entries for each license.

        Args:
            components: A list of component dictionaries. Each component should have:
                - purl: Package URL identifying the component
                - licenses: List of license dictionaries, each containing:
                    - spdxid: SPDX identifier for the license (optional)

        Returns:
            list: A list of dictionaries, each containing:
                - purl: The component's package URL
                - license: The SPDX identifier of the license (or 'Unknown' if not provided)
        """
        component_licenses: dict = {}
        for component in components:
            purl = component.get('purl', '')
            status = component.get('status', '')
            licenses = component.get('licenses', [])

            # Component without license
            if not licenses:
                key = f'{purl}-unknown'
                component_licenses[key] = {
                    'purl': purl,
                    'spdxid': 'unknown',
                    'status': status,
                    'copyleft': False,
                    'url': '-',
                }
                continue

            # Iterate over licenses component licenses
            for lic in licenses:
                spdxid = lic.get('spdxid', 'unknown')
                if spdxid not in component_licenses:
                    key = f'{purl}-{spdxid}'
                    component_licenses[key] = {
                        'purl': purl,
                        'spdxid': spdxid,
                        'status': status,
                        'copyleft': lic['copyleft'],
                        'url': lic['url'],
                    }
        return list(component_licenses.values())


#
# End of PolicyCheck Class
#



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2024, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import json
import os
import re
import unittest
from unittest.mock import Mock, patch

from src.scanoss.inspection.policy_check import PolicyStatus
from src.scanoss.inspection.raw.component_summary import ComponentSummary
from src.scanoss.inspection.raw.copyleft import Copyleft
from src.scanoss.inspection.raw.license_summary import LicenseSummary
from src.scanoss.inspection.raw.undeclared_component import UndeclaredComponent
from src.scanoss.inspection.dependency_track.project_violation import DependencyTrackProjectViolationPolicyCheck


class MyTestCase(unittest.TestCase):
    """
    Inspect for copyleft licenses
    """

    def test_copyleft_policy(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='json')
        copyleft.run()
        self.assertEqual(True, True)

    """
       Inspect for copyleft licenses empty path
    """

    def test_copyleft_policy_empty_path(self):
        copyleft = Copyleft(filepath='', format_type='json')
        success, results = copyleft.run()
        self.assertTrue(success, 2)

    """
    Inspect for empty copyleft licenses
    """

    def test_empty_copyleft_policy(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result-no-copyleft.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='json')
        status, data = copyleft.run()
        details = json.loads(data['details'])
        self.assertEqual(status, PolicyStatus.POLICY_SUCCESS.value)
        self.assertEqual(details, {})
        self.assertEqual(data['summary'], '0 component(s) with copyleft licenses were found.\n')

    """
    Inspect for copyleft licenses include
    """

    def test_copyleft_policy_include(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='json', include='MIT')
        status, data = copyleft.run()
        has_mit_license = False
        details = json.loads(data['details'])
        for component in details['components']:
            for license in component['licenses']:
                if license['spdxid'] == 'MIT':
                    has_mit_license = True
                    break

        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)
        self.assertEqual(has_mit_license, True)

    """
       Inspect for copyleft licenses exclude
    """

    def test_copyleft_policy_exclude(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='json', exclude='GPL-2.0-only')
        status, data = copyleft.run()
        results = json.loads(data['details'])
        self.assertEqual(results, {})
        self.assertEqual(status, PolicyStatus.POLICY_SUCCESS.value)

    """
        Inspect for copyleft licenses explicit
    """

    def test_copyleft_policy_explicit(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='json', explicit='MIT')
        status, data = copyleft.run()
        results = json.loads(data['details'])
        self.assertEqual(len(results['components']), 2)
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)

    """
        Inspect for copyleft licenses empty explicit licenses (should set the default ones)
    """

    def test_copyleft_policy_empty_explicit(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='json', explicit='')
        status, data = copyleft.run()
        results = json.loads(data['details'])
        self.assertEqual(len(results['components']), 5)
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)

    """
        Export copyleft licenses in Markdown
    """

    def test_copyleft_policy_markdown(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='md', explicit='MIT')
        status, data = copyleft.run()
        expected_detail_output = (
            '### Copyleft Licenses \n  | Component | License | URL | Copyleft |\n'
            ' | - | :-: | - | - |\n'
            ' | pkg:npm/%40electron/rebuild | MIT | https://spdx.org/licenses/MIT.html | YES |\n'
            '| pkg:npm/%40emotion/react | MIT | https://spdx.org/licenses/MIT.html | YES | \n'
        )
        expected_summary_output = '2 component(s) with copyleft licenses were found.\n'
        self.assertEqual(
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', data['details']),
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', expected_detail_output),
        )
        self.assertEqual(data['summary'], expected_summary_output)
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)

    ## Undeclared Components Policy Tests ##

    """
       Inspect for undeclared components empty path
    """

    def test_undeclared_policy_empty_path(self):
        undeclared = UndeclaredComponent(filepath='', format_type='json')
        success, results = undeclared.run()
        self.assertTrue(success, 2)

    """
    Inspect for undeclared components
    """

    def test_undeclared_policy(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        undeclared = UndeclaredComponent(filepath=input_file_name, format_type='json', sbom_format='legacy')
        status, data = undeclared.run()
        results = json.loads(data['details'])
        summary = data['summary']
        expected_summary_output = """3 undeclared component(s) were found.
        Add the following snippet into your `sbom.json` file 
        ```json 
        {
            "components":[
                  {
                    "purl": "pkg:github/scanoss/jenkins-pipeline-example"
                  },
                  {
                    "purl": "pkg:github/scanoss/scanner.c"
                  },
                  {
                    "purl": "pkg:github/scanoss/wfp"
                  }
            ]
        }```
        """
        self.assertEqual(len(results['components']), 4)
        self.assertEqual(
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', summary), re.sub(r'\s|\\(?!`)|\\(?=`)', '', expected_summary_output)
        )
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)

    """
       Undeclared component markdown output
    """

    def test_undeclared_policy_markdown(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        undeclared = UndeclaredComponent(filepath=input_file_name, format_type='md', sbom_format='legacy')
        status, data = undeclared.run()
        results = data['details']
        summary = data['summary']
        expected_details_output = """ ### Undeclared components
             | Component | License | 
             | - | - | 
             | pkg:github/scanoss/jenkins-pipeline-example | unknown | 
             | pkg:github/scanoss/scanner.c | GPL-2.0-only | 
             | pkg:github/scanoss/wfp | GPL-2.0-only |  """

        expected_summary_output = """3 undeclared component(s) were found.
           Add the following snippet into your `sbom.json` file 
           ```json 
               {
                "components":[
                 {
                    "purl": "pkg:github/scanoss/jenkins-pipeline-example"
                 },
                 {
                    "purl": "pkg:github/scanoss/scanner.c"
                  },
                  {
                    "purl": "pkg:github/scanoss/wfp"
                  }         
                ]             
               }```
           """
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)
        self.assertEqual(
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', results), re.sub(r'\s|\\(?!`)|\\(?=`)', '', expected_details_output)
        )
        self.assertEqual(
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', summary), re.sub(r'\s|\\(?!`)|\\(?=`)', '', expected_summary_output)
        )

    """
         Undeclared component markdown scanoss summary output
    """

    def test_undeclared_policy_markdown_scanoss_summary(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        undeclared = UndeclaredComponent(filepath=input_file_name, format_type='md')
        status, data = undeclared.run()
        results = data['details']
        summary = data['summary']
        expected_details_output = """ ### Undeclared components
               | Component | License | 
               | - | - | 
               | pkg:github/scanoss/jenkins-pipeline-example | unknown |
               | pkg:github/scanoss/scanner.c | GPL-2.0-only | 
               | pkg:github/scanoss/wfp | GPL-2.0-only | """

        expected_summary_output = """3 undeclared component(s) were found.
            Add the following snippet into your `scanoss.json` file
            
            ```json
            {
              "bom": {
                "include": [
                  {
                    "purl": "pkg:github/scanoss/jenkins-pipeline-example"
                  },
                  {
                    "purl": "pkg:github/scanoss/scanner.c"
                  },
                  {
                    "purl": "pkg:github/scanoss/wfp"
                  }
                ]
              }
            }
            ```"""
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)
        self.assertEqual(
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', results), re.sub(r'\s|\\(?!`)|\\(?=`)', '', expected_details_output)
        )
        self.assertEqual(
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', summary), re.sub(r'\s|\\(?!`)|\\(?=`)', '', expected_summary_output)
        )

    """
        Undeclared component sbom summary output
    """

    def test_undeclared_policy_scanoss_summary(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        undeclared = UndeclaredComponent(filepath=input_file_name)
        status, data = undeclared.run()
        results = json.loads(data['details'])
        summary = data['summary']
        expected_summary_output = """3 undeclared component(s) were found.
                Add the following snippet into your `scanoss.json` file

                ```json
                {
                  "bom": {
                    "include": [
                      {
                        "purl": "pkg:github/scanoss/jenkins-pipeline-example"
                      },
                      {
                        "purl": "pkg:github/scanoss/scanner.c"
                      },
                      {
                        "purl": "pkg:github/scanoss/wfp"
                      }
                    ]
                  }
                }
                ```"""
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)
        self.assertEqual(len(results['components']), 4)
        self.assertEqual(
            re.sub(r'\s|\\(?!`)|\\(?=`)', '', summary), re.sub(r'\s|\\(?!`)|\\(?=`)', '', expected_summary_output)
        )

    def test_undeclared_policy_jira_markdown_output(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        undeclared = UndeclaredComponent(filepath=input_file_name, format_type='jira_md')
        status, data = undeclared.run()
        details = data['details']
        summary = data['summary']
        expected_details_output = """|*Component*|*License*|
|pkg:github/scanoss/jenkins-pipeline-example|unknown|
|pkg:github/scanoss/scanner.c|GPL-2.0-only|
|pkg:github/scanoss/wfp|GPL-2.0-only|
"""
        expected_summary_output = """3 undeclared component(s) were found.
Add the following snippet into your `scanoss.json` file
{code:json}
{
  "bom": {
    "include": [
      {
        "purl": "pkg:github/scanoss/jenkins-pipeline-example"
      },
      {
        "purl": "pkg:github/scanoss/scanner.c"
      },
      {
        "purl": "pkg:github/scanoss/wfp"
      }
    ]
  }
}
{code}
"""
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)
        self.assertEqual(expected_details_output, details)
        self.assertEqual(summary, expected_summary_output)

    def test_copyleft_policy_jira_markdown_output(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        copyleft = Copyleft(filepath=input_file_name, format_type='jira_md')
        status, data = copyleft.run()
        results = data['details']
        expected_details_output = """### Copyleft Licenses\n|*Component*|*License*|*URL*|*Copyleft*|
|pkg:github/scanoss/scanner.c|GPL-2.0-only|https://spdx.org/licenses/GPL-2.0-only.html|YES|
|pkg:github/scanoss/engine|GPL-2.0-only|https://spdx.org/licenses/GPL-2.0-only.html|YES|
|pkg:github/scanoss/wfp|GPL-2.0-only|https://spdx.org/licenses/GPL-2.0-only.html|YES|
"""
        self.assertEqual(status, PolicyStatus.POLICY_FAIL.value)
        self.assertEqual(expected_details_output, results)

    def test_inspect_license_summary(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        i_license_summary = LicenseSummary(filepath=input_file_name)
        license_summary = i_license_summary.run()
        self.assertEqual(license_summary['detectedLicenses'], 3)
        self.assertEqual(license_summary['detectedLicensesWithCopyleft'], 1)

    def test_inspect_license_summary_with_empty_result(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'empty-result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        i_license_summary = LicenseSummary(filepath=input_file_name)
        license_summary = i_license_summary.run()
        self.assertEqual(license_summary['detectedLicenses'], 0)
        self.assertEqual(license_summary['detectedLicensesWithCopyleft'], 0)
        self.assertEqual(len(license_summary['licenses']), 0)

    def test_inspect_component_summary(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        i_component_summary = ComponentSummary(filepath=input_file_name)
        component_summary = i_component_summary.run()
        print(component_summary)
        self.assertEqual(component_summary['totalComponents'], 4)
        self.assertEqual(component_summary['undeclaredComponents'], 3)
        self.assertEqual(component_summary['declaredComponents'], 1)
        self.assertEqual(component_summary['totalFilesDetected'], 10)
        self.assertEqual(component_summary['totalFilesUndeclared'], 8)
        self.assertEqual(component_summary['totalFilesDeclared'], 2)

    def test_inspect_component_summary_empty_result(self):
        script_dir = os.path.dirname(os.path.abspath(__file__))
        file_name = 'empty-result.json'
        input_file_name = os.path.join(script_dir, 'data', file_name)
        i_component_summary = ComponentSummary(filepath=input_file_name)
        component_summary = i_component_summary.run()
        self.assertEqual(component_summary['totalComponents'], 0)
        self.assertEqual(component_summary['undeclaredComponents'], 0)
        self.assertEqual(component_summary['declaredComponents'], 0)
        self.assertEqual(len(component_summary['components']), 0)
        self.assertEqual(component_summary['totalFilesDetected'], 0)
        self.assertEqual(component_summary['totalFilesUndeclared'], 0)
        self.assertEqual(component_summary['totalFilesDeclared'], 0)

    ## Dependency Track Project Violation Policy Tests ##

    @patch('src.scanoss.inspection.dependency_track.project_violation.DependencyTrackService')
    def test_dependency_track_project_violation_json_formatter(self, mock_service):
        mock_service.return_value = Mock()
        project_violation = DependencyTrackProjectViolationPolicyCheck(
            format_type='json',
            api_key='test_key',
            url='http://localhost',
            project_id='test_project'
        )
        test_violations = [
            {
                'uuid': 'violation-1',
                'type': 'SECURITY',
                'timestamp': 1640995200000,
                'component': {
                    'name': 'test-component',
                    'version': '1.0.0',
                    'purl': 'pkg:npm/test-component@1.0.0'
                },
                'policyCondition': {
                    'policy': {
                        'name': 'Security Policy',
                        'violationState': 'FAIL'
                    }
                }
            }
        ]
        result = project_violation._json(test_violations)
        self.assertIn('details', result)
        self.assertIn('summary', result)
        self.assertEqual(result['summary'], '1 policy violations were found.\n')
        details = json.loads(result['details'])
        self.assertEqual(len(details), 1)
        self.assertEqual(details[0]['type'], 'SECURITY')

    @patch('src.scanoss.inspection.dependency_track.project_violation.DependencyTrackService')
    def test_dependency_track_project_violation_markdown_formatter(self, mock_service):
        mock_service.return_value = Mock()
        project_violation = DependencyTrackProjectViolationPolicyCheck(
            format_type='md',
            api_key='test_key',
            url='http://localhost',
            project_id='test_project'
        )
        test_violations = [
            {
                'uuid': 'violation-1',
                'type': 'SECURITY',
                'timestamp': 1640995200000,
                'component': {
                    'name': 'test-component',
                    'version': '1.0.0',
                    'purl': 'pkg:npm/test-component@1.0.0'
                },
                'policyCondition': {
                    'policy': {
                        'name': 'Security Policy',
                        'violationState': 'FAIL'
                    }
                }
            }
        ]
        result = project_violation._markdown(test_violations)
        self.assertIn('details', result)
        self.assertIn('summary', result)
        self.assertEqual(result['summary'], '1 policy violations were found.\n')
        self.assertIn('State', result['details'])
        self.assertIn('Risk Type', result['details'])
        self.assertIn('Policy Name', result['details'])
        self.assertIn('Component', result['details'])
        self.assertIn('Date', result['details'])

    @patch('src.scanoss.inspection.dependency_track.project_violation.DependencyTrackService')
    def test_dependency_track_project_violation_sort_violations(self, mock_service):
        mock_service.return_value = Mock()
        project_violation = DependencyTrackProjectViolationPolicyCheck(
            api_key='test_key',
            url='http://localhost',
            project_id='test_project'
        )
        test_violations = [
            {'type': 'LICENSE', 'uuid': 'license-violation'},
            {'type': 'SECURITY', 'uuid': 'security-violation'},
            {'type': 'OTHER', 'uuid': 'other-violation'},
            {'type': 'SECURITY', 'uuid': 'security-violation-2'}
        ]
        sorted_violations = project_violation._sort_project_violations(test_violations)
        self.assertEqual(sorted_violations[0]['type'], 'SECURITY')
        self.assertEqual(sorted_violations[1]['type'], 'SECURITY')
        self.assertEqual(sorted_violations[2]['type'], 'LICENSE')
        self.assertEqual(sorted_violations[3]['type'], 'OTHER')

    @patch('src.scanoss.inspection.dependency_track.project_violation.DependencyTrackService')
    def test_dependency_track_project_violation_empty_violations(self, mock_service):
        mock_service.return_value = Mock()
        project_violation = DependencyTrackProjectViolationPolicyCheck(
            format_type='json',
            api_key='test_key',
            url='http://localhost',
            project_id='test_project'
        )
        empty_violations = []
        result = project_violation._json(empty_violations)
        self.assertEqual(result['summary'], '0 policy violations were found.\n')
        details = json.loads(result['details'])
        self.assertEqual(len(details), 0)

    @patch('src.scanoss.inspection.dependency_track.project_violation.DependencyTrackService')
    def test_dependency_track_project_violation_markdown_empty(self, mock_service):
        mock_service.return_value = Mock()
        project_violation = DependencyTrackProjectViolationPolicyCheck(
            format_type='md',
            api_key='test_key',
            url='http://localhost',
            project_id='test_project'
        )
        empty_violations = []
        result = project_violation._markdown(empty_violations)
        self.assertEqual(result['summary'], '0 policy violations were found.\n')
        self.assertIn('State', result['details'])
        self.assertIn('Risk Type', result['details'])

    @patch('src.scanoss.inspection.dependency_track.project_violation.DependencyTrackService')
    def test_dependency_track_project_violation_multiple_types(self, mock_service):
        mock_service.return_value = Mock()
        project_violation = DependencyTrackProjectViolationPolicyCheck(
            format_type='json',
            api_key='test_key',
            url='http://localhost',
            project_id='test_project'
        )
        test_violations = [
            {
                'uuid': 'violation-1',
                'type': 'SECURITY',
                'timestamp': 1640995200000,
                'component': {
                    'name': 'vulnerable-component',
                    'version': '1.0.0',
                    'purl': 'pkg:npm/vulnerable-component@1.0.0'
                },
                'policyCondition': {
                    'policy': {
                        'name': 'Security Policy',
                        'violationState': 'FAIL'
                    }
                }
            },
            {
                'uuid': 'violation-2',
                'type': 'LICENSE',
                'timestamp': 1640995300000,
                'component': {
                    'name': 'license-component',
                    'version': '2.0.0',
                    'purl': 'pkg:npm/license-component@2.0.0'
                },
                'policyCondition': {
                    'policy': {
                        'name': 'License Policy',
                        'violationState': 'WARN'
                    }
                }
            }
        ]
        result = project_violation._json(test_violations)
        self.assertEqual(result['summary'], '2 policy violations were found.\n')
        details = json.loads(result['details'])
        self.assertEqual(len(details), 2)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: scanoss/api/scanning/v2/scanoss-scanning.proto
# Protobuf Python Version: 6.31.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    0,
    '',
    'scanoss/api/scanning/v2/scanoss-scanning.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from scanoss.api.common.v2 import scanoss_common_pb2 as scanoss_dot_api_dot_common_dot_v2_dot_scanoss__common__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from protoc_gen_openapiv2.options import annotations_pb2 as protoc__gen__openapiv2_dot_options_dot_annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.scanoss/api/scanning/v2/scanoss-scanning.proto\x12\x17scanoss.api.scanning.v2\x1a*scanoss/api/common/v2/scanoss-common.proto\x1a\x1cgoogle/api/annotations.proto\x1a.protoc-gen-openapiv2/options/annotations.proto\"\xc5\x03\n\nHFHRequest\x12:\n\x04root\x18\x01 \x01(\x0b\x32,.scanoss.api.scanning.v2.HFHRequest.Children\x12\x16\n\x0erank_threshold\x18\x02 \x01(\x05\x12\x10\n\x08\x63\x61tegory\x18\x03 \x01(\t\x12\x13\n\x0bquery_limit\x18\x04 \x01(\x05\x1a\xbb\x02\n\x08\x43hildren\x12\x0f\n\x07path_id\x18\x01 \x01(\t\x12\x16\n\x0esim_hash_names\x18\x02 \x01(\t\x12\x18\n\x10sim_hash_content\x18\x03 \x01(\t\x12>\n\x08\x63hildren\x18\x04 \x03(\x0b\x32,.scanoss.api.scanning.v2.HFHRequest.Children\x12\x1a\n\x12sim_hash_dir_names\x18\x05 \x01(\t\x12Y\n\x0flang_extensions\x18\x06 \x03(\x0b\x32@.scanoss.api.scanning.v2.HFHRequest.Children.LangExtensionsEntry\x1a\x35\n\x13LangExtensionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"\xa3\x03\n\x0bHFHResponse\x12<\n\x07results\x18\x01 \x03(\x0b\x32+.scanoss.api.scanning.v2.HFHResponse.Result\x12\x35\n\x06status\x18\x02 \x01(\x0b\x32%.scanoss.api.common.v2.StatusResponse\x1a)\n\x07Version\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\r\n\x05score\x18\x02 \x01(\x02\x1a\x94\x01\n\tComponent\x12\x0c\n\x04purl\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06vendor\x18\x03 \x01(\t\x12>\n\x08versions\x18\x04 \x03(\x0b\x32,.scanoss.api.scanning.v2.HFHResponse.Version\x12\x0c\n\x04rank\x18\x05 \x01(\x05\x12\r\n\x05order\x18\x06 \x01(\x05\x1a]\n\x06Result\x12\x0f\n\x07path_id\x18\x01 \x01(\t\x12\x42\n\ncomponents\x18\x02 \x03(\x0b\x32..scanoss.api.scanning.v2.HFHResponse.Component2\x81\x02\n\x08Scanning\x12q\n\x04\x45\x63ho\x12\".scanoss.api.common.v2.EchoRequest\x1a#.scanoss.api.common.v2.EchoResponse\" \x82\xd3\xe4\x93\x02\x1a\"\x15/api/v2/scanning/echo:\x01*\x12\x81\x01\n\x0e\x46olderHashScan\x12#.scanoss.api.scanning.v2.HFHRequest\x1a$.scanoss.api.scanning.v2.HFHResponse\"$\x82\xd3\xe4\x93\x02\x1e\"\x19/api/v2/scanning/hfh/scan:\x01*B\x8a\x02Z1github.com/scanoss/papi/api/scanningv2;scanningv2\x92\x41\xd3\x01\x12m\n\x18SCANOSS Scanning Service\"L\n\x10scanoss-scanning\x12#https://github.com/scanoss/scanning\x1a\x13support@scanoss.com2\x03\x32.0*\x01\x01\x32\x10\x61pplication/json:\x10\x61pplication/jsonR;\n\x03\x34\x30\x34\x12\x34\n*Returned when the resource does not exist.\x12\x06\n\x04\x9a\x02\x01\x07\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'scanoss.api.scanning.v2.scanoss_scanning_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z1github.com/scanoss/papi/api/scanningv2;scanningv2\222A\323\001\022m\n\030SCANOSS Scanning Service\"L\n\020scanoss-scanning\022#https://github.com/scanoss/scanning\032\023support@scanoss.com2\0032.0*\001\0012\020application/json:\020application/jsonR;\n\003404\0224\n*Returned when the resource does not exist.\022\006\n\004\232\002\001\007'
  _globals['_HFHREQUEST_CHILDREN_LANGEXTENSIONSENTRY']._loaded_options = None
  _globals['_HFHREQUEST_CHILDREN_LANGEXTENSIONSENTRY']._serialized_options = b'8\001'
  _globals['_SCANNING'].methods_by_name['Echo']._loaded_options = None
  _globals['_SCANNING'].methods_by_name['Echo']._serialized_options = b'\202\323\344\223\002\032\"\025/api/v2/scanning/echo:\001*'
  _globals['_SCANNING'].methods_by_name['FolderHashScan']._loaded_options = None
  _globals['_SCANNING'].methods_by_name['FolderHashScan']._serialized_options = b'\202\323\344\223\002\036\"\031/api/v2/scanning/hfh/scan:\001*'
  _globals['_HFHREQUEST']._serialized_start=198
  _globals['_HFHREQUEST']._serialized_end=651
  _globals['_HFHREQUEST_CHILDREN']._serialized_start=336
  _globals['_HFHREQUEST_CHILDREN']._serialized_end=651
  _globals['_HFHREQUEST_CHILDREN_LANGEXTENSIONSENTRY']._serialized_start=598
  _globals['_HFHREQUEST_CHILDREN_LANGEXTENSIONSENTRY']._serialized_end=651
  _globals['_HFHRESPONSE']._serialized_start=654
  _globals['_HFHRESPONSE']._serialized_end=1073
  _globals['_HFHRESPONSE_VERSION']._serialized_start=786
  _globals['_HFHRESPONSE_VERSION']._serialized_end=827
  _globals['_HFHRESPONSE_COMPONENT']._serialized_start=830
  _globals['_HFHRESPONSE_COMPONENT']._serialized_end=978
  _globals['_HFHRESPONSE_RESULT']._serialized_start=980
  _globals['_HFHRESPONSE_RESULT']._serialized_end=1073
  _globals['_SCANNING']._serialized_start=1076
  _globals['_SCANNING']._serialized_end=1333
# @@protoc_insertion_point(module_scope)



DEFAULT_POST_SIZE = 32
DEFAULT_TIMEOUT = 180
DEFAULT_RETRY = 5
MIN_TIMEOUT = 5

PYTHON_MAJOR_VERSION = 3

DEFAULT_SC_TIMEOUT = 600
DEFAULT_NB_THREADS = 5

DEFAULT_URL = 'https://api.osskb.org'  # default free service URL
DEFAULT_URL2 = 'https://api.scanoss.com'  # default premium service URL

DEFAULT_API_TIMEOUT = 600

DEFAULT_HFH_RANK_THRESHOLD = 5


"""
SPDX-License-Identifier: MIT

  Copyright (c) 2021, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import os
import queue
import sys
import threading
import time
from dataclasses import dataclass
from typing import Dict, List

from progress.bar import Bar

from .scanossapi import ScanossApi
from .scanossbase import ScanossBase

WFP_FILE_START = 'file='
MAX_ALLOWED_THREADS = (
    int(os.environ.get('SCANOSS_MAX_ALLOWED_THREADS')) if os.environ.get('SCANOSS_MAX_ALLOWED_THREADS') else 30
)


@dataclass
class ThreadedScanning(ScanossBase):
    """
    Threaded class for running Scanning in parallel (from a queue)
    WFP scan requests are loaded into the input queue.
    Multiple threads pull messages off this queue, process the request and put the results into an output queue
    """

    bar: Bar = None

    def __init__(
        self, scanapi: ScanossApi, debug: bool = False, trace: bool = False, quiet: bool = False, nb_threads: int = 5
    ) -> None:
        """
        Initialise the ThreadedScanning class
        :param scanapi: SCANOSS API to send scan requests to
        :param debug: enable debug (default False)
        :param trace: enable trace (default False)
        :param quiet: enable quiet mode (default False)
        :param nb_threads: Number of thread to run (default 5)
        """
        super().__init__(debug, trace, quiet)
        self.inputs = queue.Queue()
        self.output = queue.Queue()
        self.scanapi = scanapi
        self.nb_threads = nb_threads
        self._isatty = sys.stderr.isatty()
        self._bar_count = 0
        self._errors = False
        self._lock = threading.Lock()
        self._stop_event = threading.Event()  # Control when scanning threads should terminate
        self._stop_scanning = threading.Event()  # Control if the parent process should abort scanning
        self._threads = []
        if nb_threads > MAX_ALLOWED_THREADS:
            self.print_msg(f'Warning: Requested threads too large: {nb_threads}. Reducing to {MAX_ALLOWED_THREADS}')
            self.nb_threads = MAX_ALLOWED_THREADS

    @staticmethod
    def __count_files_in_wfp(wfp: str):
        """
        Count the number of files in the WFP that need to be processed
        :param wfp: WFP string
        :return: number of files in the WFP
        """
        count = 0
        if wfp:
            for line in wfp.split('\n'):
                if WFP_FILE_START in line:
                    count += 1
        return count

    def create_bar(self, file_count: int):
        if not self.quiet and self._isatty and not self.bar:
            self.bar = Bar('Scanning', max=file_count)
            self.bar.next(self._bar_count)

    def complete_bar(self):
        if self.bar:
            self.bar.finish()

    def set_bar(self, bar: Bar) -> None:
        """
        Set the Progress Bar to display progress while scanning
        :param bar: Progress Bar object
        """
        self.bar = bar

    def update_bar(self, amount: int = 0, create: bool = False, file_count: int = 0) -> None:
        """
        Update the Progress Bar progress
        :param amount: amount of progress to update
        :param create: create the bar if requested
        :param file_count: file count
        """
        try:
            self._lock.acquire()
            try:
                if create and not self.bar:
                    self.create_bar(file_count)
                elif self.bar:
                    self.bar.next(amount)
                self._bar_count += amount
            finally:
                self._lock.release()
        except Exception as e:
            self.print_debug(f'Warning: Update status bar lock failed: {e}. Ignoring.')

    def queue_add(self, wfp: str) -> None:
        """
        Add requests to the queue
        :param wfp: WFP to add to queue
        """
        if wfp is None or wfp == '':
            self.print_stderr('Warning: empty WFP. Skipping from scan...')
        else:
            self.inputs.put(wfp)

    def get_queue_size(self) -> int:
        return self.inputs.qsize()

    def stop_scanning(self) -> bool:
        """
        Check if we should keep scanning or not
        """
        return self._stop_scanning.is_set()

    @property
    def responses(self) -> List[Dict]:
        """
        Get all responses back from the completed threads
        :return: List of JSON objects
        """
        return list(self.output.queue)

    def run(self, wait: bool = True) -> bool:
        """
        Initiate the threads and process all pending requests
        :return: True if successful, False if error encountered
        """
        qsize = self.inputs.qsize()
        if qsize < self.nb_threads:
            self.print_debug(
                f'Input queue ({qsize}) smaller than requested threads: {self.nb_threads}. Reducing to queue size.'
            )
            self.nb_threads = qsize
        else:
            self.print_debug(f'Starting {self.nb_threads} threads to process {qsize} requests...')
        try:
            for i in range(0, self.nb_threads):
                t = threading.Thread(target=self.worker_post, daemon=True)
                self._threads.append(t)
                t.start()
        except Exception as e:
            self.print_stderr(f'ERROR: Problem running threaded scanning: {e}')
            self._errors = True
        if wait:  # Wait for all inputs to complete
            self.complete()
        return False if self._errors else True

    def complete(self) -> bool:
        """
        Wait for input queue to complete processing and complete the worker threads
        """
        self.inputs.join()
        self._stop_event.set()  # Tell the worker threads to stop
        try:
            for t in self._threads:  # Complete the threads
                t.join(timeout=5)
        except Exception as e:
            self.print_stderr(f'WARNING: Issue encountered terminating scanning worker threads: {e}')
            self._errors = True
        return False if self._errors else True

    def worker_post(self) -> None:
        """
        Take each request and process it
        :return: None
        """
        current_thread = threading.get_ident()
        self.print_trace(f'Starting worker {current_thread}...')
        api_error = False
        while not self._stop_event.is_set():
            wfp = None
            if not self.inputs.empty():  # Only try to get a message if there is one on the queue
                try:
                    wfp = self.inputs.get(timeout=5)
                    if api_error:  # API error encountered, so stop processing anymore requests
                        self.inputs.task_done()  # remove request from the queue
                    else:
                        self.print_trace(f'Processing input request ({current_thread})...')
                        count = self.__count_files_in_wfp(wfp)
                        if wfp is None or wfp == '':
                            self.print_stderr(f'Warning: Empty WFP in request input: {wfp}')
                        resp = self.scanapi.scan(wfp, scan_id=current_thread)
                        if resp:
                            self.output.put(resp)  # Store the output response to later collection
                        self.update_bar(count)
                        self.inputs.task_done()
                        self.print_trace(f'Request complete ({current_thread}).')
                except queue.Empty:
                    self.print_stderr(f'No message available to process ({current_thread}). Checking again...')
                except Exception as e:
                    self.print_stderr(f'ERROR: Problem encountered running scan: {e}. Aborting current thread.')
                    self._errors = True
                    if wfp:
                        self.inputs.task_done()  # If there was a WFP being processed, remove it from the queue
                    api_error = True  # Stop processing anymore work requests
                    self._stop_scanning.set()  # Tell the parent process to abort scanning
            else:
                time.sleep(1)  # Sleep while waiting for the queue depth to build up
        self.print_trace(f'Thread complete ({current_thread}).')


#
# End of ThreadedScanning Class
#



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2024, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

from ...scanossbase import ScanossBase

DEFAULT_COPYLEFT_LICENSES = {
    'agpl-3.0-only',
    'artistic-1.0',
    'artistic-2.0',
    'cc-by-sa-4.0',
    'cddl-1.0',
    'cddl-1.1',
    'cecill-2.1',
    'epl-1.0',
    'epl-2.0',
    'gfdl-1.1-only',
    'gfdl-1.2-only',
    'gfdl-1.3-only',
    'gpl-1.0-only',
    'gpl-2.0-only',
    'gpl-3.0-only',
    'lgpl-2.1-only',
    'lgpl-3.0-only',
    'mpl-1.1',
    'mpl-2.0',
    'sleepycat',
    'watcom-1.0',
}


class LicenseUtil(ScanossBase):
    """
    A utility class for handling software licenses, particularly copyleft licenses.

    This class provides functionality to initialize, manage, and query a set of
    copyleft licenses. It also offers a method to generate URLs for license information.
    """

    BASE_SPDX_ORG_URL = 'https://spdx.org/licenses'
    BASE_OSADL_URL = 'https://www.osadl.org/fileadmin/checklists/unreflicenses'

    def __init__(self, debug: bool = False, trace: bool = True, quiet: bool = False):
        super().__init__(debug, trace, quiet)
        self.default_copyleft_licenses = set(DEFAULT_COPYLEFT_LICENSES)
        self.copyleft_licenses = set()

    def init(self, include: str = None, exclude: str = None, explicit: str = None):
        """
        Initialize the set of copyleft licenses based on user input.

        This method allows for customization of the copyleft license set by:
        - Setting an explicit list of licenses
        - Including additional licenses to the default set
        - Excluding specific licenses from the default set

        :param include: Comma-separated string of licenses to include
        :param exclude: Comma-separated string of licenses to exclude
        :param explicit: Comma-separated string of licenses to use exclusively
        """
        if self.debug:
            self.print_stderr(f'Include Copyleft licenses: ${include}')
            self.print_stderr(f'Exclude Copyleft licenses: ${exclude}')
            self.print_stderr(f'Explicit Copyleft licenses: ${explicit}')
        if explicit:
            explicit = explicit.strip()
        if explicit:
            exp = [item.strip().lower() for item in explicit.split(',')]
            self.copyleft_licenses = set(exp)
            self.print_debug(f'Copyleft licenses: ${self.copyleft_licenses}')
            return
        # If no explicit licenses were set, set default ones
        self.copyleft_licenses = self.default_copyleft_licenses.copy()
        if include:
            include = include.strip()
        if include:
            inc = [item.strip().lower() for item in include.split(',')]
            self.copyleft_licenses.update(inc)
        if exclude:
            exclude = exclude.strip()
        if exclude:
            inc = [item.strip().lower() for item in exclude.split(',')]
            for lic in inc:
                self.copyleft_licenses.discard(lic)
        self.print_debug(f'Copyleft licenses: ${self.copyleft_licenses}')

    def is_copyleft(self, spdxid: str) -> bool:
        """
        Check if a given license is considered copyleft.

        :param spdxid: The SPDX identifier of the license to check
        :return: True if the license is copyleft, False otherwise
        """
        return spdxid.lower() in self.copyleft_licenses

    def get_spdx_url(self, spdxid: str) -> str:
        """
        Generate the URL for the SPDX page of a license.

        :param spdxid: The SPDX identifier of the license
        :return: The URL of the SPDX page for the given license
        """
        return f'{self.BASE_SPDX_ORG_URL}/{spdxid}.html'

    def get_osadl_url(self, spdxid: str) -> str:
        """
        Generate the URL for the OSADL (Open Source Automation Development Lab) page of a license.

        :param spdxid: The SPDX identifier of the license
        :return: The URL of the OSADL page for the given license
        """
        return f'{self.BASE_OSADL_URL}/{spdxid}.txt'


#
# End of LicenseUtil Class
#



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2025, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

from abc import abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, Generic, List, TypeVar

from ..scanossbase import ScanossBase
from .utils.license_utils import LicenseUtil


class PolicyStatus(Enum):
    """
    Enumeration representing the status of a policy check.

    Attributes:
        POLICY_SUCCESS (int): Indicates that the policy check passed successfully (value: 0).
        POLICY_FAIL (int): Indicates that the policy check failed (value: 2).
        ERROR (int): Indicates that an error occurred during the policy check (value: 1).
    """
    POLICY_SUCCESS = 0
    POLICY_FAIL = 2
    ERROR = 1
#
# End of PolicyStatus Class
#

T = TypeVar('T')

class PolicyCheck(ScanossBase, Generic[T]):
    """
    A base class for implementing various software policy checks.

    This class provides a framework for policy checking, including methods for
    processing components, generating output in different formats.

    Attributes:
        VALID_FORMATS (set): A set of valid output formats ('md', 'json').

    Inherits from:
        InspectBase: A base class providing common functionality for SCANOSS-related operations.
    """
    VALID_FORMATS = {'md', 'json', 'jira_md'}
    def __init__(  # noqa: PLR0913
            self,
            debug: bool = False,
            trace: bool = False,
            quiet: bool = False,
            format_type: str = None,
            status: str = None,
            name: str = None,
            output: str = None,
    ):
        super().__init__(debug, trace, quiet)
        self.license_util = LicenseUtil()
        self.name = name
        self.format_type = format_type
        self.status = status
        self.output = output

    @abstractmethod
    def run(self):
        """
        Execute the policy check process.

        This abstract method should be implemented by subclasses to perform specific
        policy checks. The general structure of this method typically includes:
        1. Retrieving components
        2. Filtering components based on specific criteria
        3. Formatting the results
        4. Saving the output to files if required

        :return: A tuple containing:
                 - First element: PolicyStatus enum value (SUCCESS, FAIL, or ERROR)
                 - Second element: Dictionary containing the inspection results
        """
        pass

    @abstractmethod
    def _json(self, data: list[T]) -> Dict[str, Any]:
        """
        Format the policy checks results as JSON.
        This method should be implemented by subclasses to create a Markdown representation
        of the policy check results.

        :param data: List of data to be formatted.
        :return: A dictionary containing two keys:
                 - 'results': A JSON-formatted string with the full list of components
                 - 'summary': A string summarizing the number of components found
        """
        pass

    @abstractmethod
    def _markdown(self, data: list[T]) -> Dict[str, Any]:
        """
        Generate Markdown output for the policy check results.

        This method should be implemented by subclasses to create a Markdown representation
        of the policy check results.

        :param data: List of data to be included in the output.
        :return: A dictionary representing the Markdown output.
        """
        pass

    @abstractmethod
    def _jira_markdown(self, data: list[T]) -> Dict[str, Any]:
        """
        Generate Markdown output for the policy check results.

        This method should be implemented by subclasses to create a Markdown representation
        of the policy check results.

        :param data: List of data to be included in the output.
        :return: A dictionary representing the Markdown output.
        """
        pass

    def generate_table(self, headers, rows, centered_columns=None):
        """
        Generate a Markdown table.

        :param headers: List of headers for the table.
        :param rows: List of rows for the table.
        :param centered_columns: List of column indices to be centered.
        :return: A string representing the Markdown table.
        """
        col_sep = ' | '
        centered_column_set = set(centered_columns or [])
        if headers is None:
            self.print_stderr('ERROR: Header are no set')
            return None

        # Decide which separator to use
        def create_separator(index):
            if centered_columns is None:
                return '-'
            return ':-:' if index in centered_column_set else '-'

        # Build the row separator
        row_separator = col_sep + col_sep.join(create_separator(index) for index, _ in enumerate(headers)) + col_sep
        # build table rows
        table_rows = [col_sep + col_sep.join(headers) + col_sep, row_separator]
        table_rows.extend(col_sep + col_sep.join(row) + col_sep for row in rows)
        return '\n'.join(table_rows)

    def generate_jira_table(self, headers, rows, centered_columns=None):
        col_sep = '*|*'
        if headers is None:
            self.print_stderr('ERROR: Header are no set')
            return None

        table_header = '|*' + col_sep.join(headers) + '*|\n'
        table = table_header
        for row in rows:
            if len(headers) == len(row):
                table += '|' + '|'.join(row) + '|\n'

        return table

    def _get_formatter(self) -> Callable[[List[dict]], Dict[str, Any]] or None:
        """
        Get the appropriate formatter function based on the specified format.

        :return: Formatter function (either _json or _markdown)
        """
        valid_format = self._is_valid_format()
        if not valid_format:
            return None
        # a map of which format function to return
        function_map = {
            'json': self._json,
            'md': self._markdown,
            'jira_md': self._jira_markdown,
        }
        return function_map[self.format_type]

    def _debug(self):
        """
        Print debug information about the policy check.

        This method prints various attributes of the PolicyCheck instance for debugging purposes.
        """
        if self.debug:
            self.print_stderr(f'Policy: {self.name}')
            self.print_stderr(f'Format: {self.format_type}')
            self.print_stderr(f'Status: {self.status}')
            self.print_stderr(f'Output: {self.output}')

    def _is_valid_format(self) -> bool:
        """
        Validate if the format specified is supported.

        This method checks if the format stored in format is one of the
        valid formats defined in self.VALID_FORMATS.

        :return: bool: True if the format is valid, False otherwise.
        """
        if self.format_type not in self.VALID_FORMATS:
            valid_formats_str = ', '.join(self.VALID_FORMATS)
            self.print_stderr(f'ERROR: Invalid format "{self.format_type}". Valid formats are: {valid_formats_str}')
            return False
        return True

    def _generate_formatter_report(self, components: list[Dict]):
        """
        Generates a formatted report for a given component based on the defined formatter.

        Parameters:
            components (List[dict]): A list of dictionaries representing the components to be
            processed and formatted. Each dictionary contains detailed information that adheres
            to the format requirements for the specified formatter.

        Returns:
            Tuple[int, dict]: A tuple where the first element represents the policy status code
            and the second element is a dictionary containing formatted results information,
            typically with keys 'details' and 'summary'.

        Raises:
            KeyError: When a required key is missing from the provided component, causing the
            formatter to fail.
            ValueError: If an invalid component is passed and renders unable to process.
        """
        # Get a formatter for the output results
        formatter = self._get_formatter()
        if formatter is None:
            return PolicyStatus.ERROR.value, {}
        # Format the results
        data = formatter(components)
        ## Save outputs if required
        self.print_to_file_or_stdout(data['details'], self.output)
        self.print_to_file_or_stderr(data['summary'], self.status)
        # Check to see if we have policy violations
        if len(components) > 0:
            return PolicyStatus.POLICY_FAIL.value, data
        return PolicyStatus.POLICY_SUCCESS.value, data
#
# End of PolicyCheck Class
#


from abc import ABC, abstractmethod

from scanoss.scanossbase import ScanossBase


class AbstractPresenter(ABC):
    """
    Abstract presenter class for presenting output in a given format.
    Subclasses must implement the _format_json_output and _format_plain_output methods.
    """

    def __init__(
        self,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
        output_file: str = None,
        output_format: str = None,
    ):
        """
        Initialize the presenter with the given output file and format.
        """
        self.AVAILABLE_OUTPUT_FORMATS = ['json', 'plain', 'cyclonedx', 'spdxlite', 'csv', 'raw']
        self.base = ScanossBase(debug=debug, trace=trace, quiet=quiet)
        self.output_file = output_file
        self.output_format = output_format

    def present(self, output_format: str = None, output_file: str = None):
        """
        Present the formatted output to a file if provided; otherwise, print to stdout.
        """
        file_path = output_file or self.output_file
        fmt = output_format or self.output_format

        if fmt and fmt not in self.AVAILABLE_OUTPUT_FORMATS:
            raise ValueError(
                f"ERROR: Invalid output format '{fmt}'. Valid values are: {', '.join(self.AVAILABLE_OUTPUT_FORMATS)}"
            )

        if fmt == 'json':
            content = self._format_json_output()
        elif fmt == 'plain':
            content = self._format_plain_output()
        elif fmt == 'cyclonedx':
            content = self._format_cyclonedx_output()
        elif fmt == 'spdxlite':
            content = self._format_spdxlite_output()
        elif fmt == 'csv':
            content = self._format_csv_output()
        elif fmt == 'raw':
            content = self._format_raw_output()
        else:
            content = self._format_plain_output()

        self._present_output(content, file_path)

    def _present_output(self, content: str, file_path: str = None):
        """
        If a file path is provided, write to that file; otherwise, print the content to stdout.
        """
        self.base.print_to_file_or_stdout(content, file_path)

    @abstractmethod
    def _format_cyclonedx_output(self) -> str:
        """
        Return a CycloneDX string representation of the data.
        """
        pass

    @abstractmethod
    def _format_spdxlite_output(self) -> str:
        """
        Return a SPDX-Lite string representation of the data.
        """
        pass

    @abstractmethod
    def _format_csv_output(self) -> str:
        """
        Return a CSV string representation of the data.
        """
        pass

    @abstractmethod
    def _format_json_output(self) -> str:
        """
        Return a JSON string representation of the data.
        """
        pass

    @abstractmethod
    def _format_plain_output(self) -> str:
        """
        Return a plain text string representation of the data.
        """
        pass

    @abstractmethod
    def _format_raw_output(self) -> str:
        """
        Return a raw string representation of the data.
        """
        pass



"""
SPDX-License-Identifier: BSD-3-Clause

  Copyright (c) 2015, Gengo, Inc.
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification,
  are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   * Neither the name of Gengo, Inc. nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.
"""



"""
SPDX-License-Identifier: MIT

  Copyright (c) 2021, SCANOSS

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
"""

import concurrent.futures
import json
import os
import uuid
from dataclasses import dataclass
from enum import IntEnum
from typing import Dict, Optional
from urllib.parse import urlparse

import grpc
from google.protobuf.json_format import MessageToDict, ParseDict
from pypac.parser import PACFile
from pypac.resolver import ProxyResolver

from scanoss.api.scanning.v2.scanoss_scanning_pb2_grpc import ScanningStub
from scanoss.constants import DEFAULT_TIMEOUT

from . import __version__
from .api.common.v2.scanoss_common_pb2 import (
    EchoRequest,
    EchoResponse,
    PurlRequest,
    StatusCode,
    StatusResponse,
)
from .api.components.v2.scanoss_components_pb2 import (
    CompSearchRequest,
    CompSearchResponse,
    CompVersionRequest,
    CompVersionResponse,
)
from .api.components.v2.scanoss_components_pb2_grpc import ComponentsStub
from .api.cryptography.v2.scanoss_cryptography_pb2_grpc import CryptographyStub
from .api.dependencies.v2.scanoss_dependencies_pb2 import DependencyRequest
from .api.dependencies.v2.scanoss_dependencies_pb2_grpc import DependenciesStub
from .api.geoprovenance.v2.scanoss_geoprovenance_pb2 import ContributorResponse
from .api.geoprovenance.v2.scanoss_geoprovenance_pb2_grpc import GeoProvenanceStub
from .api.scanning.v2.scanoss_scanning_pb2 import HFHRequest
from .api.semgrep.v2.scanoss_semgrep_pb2 import SemgrepResponse
from .api.semgrep.v2.scanoss_semgrep_pb2_grpc import SemgrepStub
from .api.vulnerabilities.v2.scanoss_vulnerabilities_pb2 import VulnerabilityResponse
from .api.vulnerabilities.v2.scanoss_vulnerabilities_pb2_grpc import VulnerabilitiesStub
from .scanossbase import ScanossBase

DEFAULT_URL = 'https://api.osskb.org'  # default free service URL
DEFAULT_URL2 = 'https://api.scanoss.com'  # default premium service URL
SCANOSS_GRPC_URL = os.environ.get('SCANOSS_GRPC_URL') if os.environ.get('SCANOSS_GRPC_URL') else DEFAULT_URL
SCANOSS_API_KEY = os.environ.get('SCANOSS_API_KEY') if os.environ.get('SCANOSS_API_KEY') else ''

MAX_CONCURRENT_REQUESTS = 5


class ScanossGrpcError(Exception):
    """
    Custom exception for SCANOSS gRPC errors
    """

    pass


class ScanossGrpcStatusCode(IntEnum):
    """Status codes for SCANOSS gRPC responses"""

    SUCCESS = 1
    SUCCESS_WITH_WARNINGS = 2
    FAILED_WITH_WARNINGS = 3
    FAILED = 4


class ScanossGrpc(ScanossBase):
    """
    Client for gRPC functionality
    """

    def __init__(  # noqa: PLR0913, PLR0915
        self,
        url: str = None,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
        ca_cert: str = None,
        api_key: str = None,
        ver_details: str = None,
        timeout: int = 600,
        proxy: str = None,
        grpc_proxy: str = None,
        pac: PACFile = None,
        req_headers: dict = None,
    ):
        """

        :param url:
        :param debug:
        :param trace:
        :param quiet:
        :param ca_cert:

        To set a custom certificate use:
            GRPC_DEFAULT_SSL_ROOTS_FILE_PATH=/path/to/certs/cert.pem
        More details here: https://grpc.github.io/grpc/cpp/grpc__security__constants_8h.html
                           https://github.com/grpc/grpc/blob/master/doc/environment_variables.md
        To enable a Proxy use:
            grpc_proxy='http://<ip>:<port>'
        """
        super().__init__(debug, trace, quiet)
        self.url = url
        self.api_key = api_key if api_key else SCANOSS_API_KEY
        self.timeout = timeout
        self.proxy = proxy
        self.grpc_proxy = grpc_proxy
        self.pac = pac
        self.req_headers = req_headers
        self.metadata = []

        if self.api_key:
            self.metadata.append(('x-api-key', api_key))  # Set API key if we have one
        if ver_details:
            self.metadata.append(('x-scanoss-client', ver_details))
        self.metadata.append(('user-agent', f'scanoss-py/{__version__}'))
        self.load_generic_headers()

        self.url = url if url else SCANOSS_GRPC_URL
        if self.api_key and not url and not os.environ.get('SCANOSS_GRPC_URL'):
            self.url = DEFAULT_URL2  # API key specific and no alternative URL, so use the default premium
        self.url = self.url.lower()
        self.orig_url = self.url  # Used for proxy lookup

        secure = True if self.url.startswith('https:') else False  # Is it a secure connection?
        if self.url.startswith('http'):
            u = urlparse(self.url)
            port = u.port
            if port is None:
                port = 443 if u.scheme == 'https' else 80  # Set the default port number if it's not available
            self.url = f'{u.hostname}:{port}'
        cert_data = None
        if ca_cert is not None:
            secure = True
            cert_data = ScanossGrpc._load_cert(ca_cert)
        self.print_debug(f'Setting up (secure: {secure}) connection to {self.url}...')
        self._get_proxy_config()
        if secure is False:  # insecure connection
            self.comp_search_stub = ComponentsStub(grpc.insecure_channel(self.url))
            self.crypto_stub = CryptographyStub(grpc.insecure_channel(self.url))
            self.dependencies_stub = DependenciesStub(grpc.insecure_channel(self.url))
            self.semgrep_stub = SemgrepStub(grpc.insecure_channel(self.url))
            self.vuln_stub = VulnerabilitiesStub(grpc.insecure_channel(self.url))
            self.provenance_stub = GeoProvenanceStub(grpc.insecure_channel(self.url))
            self.scanning_stub = ScanningStub(grpc.insecure_channel(self.url))
        else:
            if ca_cert is not None:
                credentials = grpc.ssl_channel_credentials(cert_data)  # secure with specified certificate
            else:
                credentials = grpc.ssl_channel_credentials()  # secure connection with default certificate
            self.comp_search_stub = ComponentsStub(grpc.secure_channel(self.url, credentials))
            self.crypto_stub = CryptographyStub(grpc.secure_channel(self.url, credentials))
            self.dependencies_stub = DependenciesStub(grpc.secure_channel(self.url, credentials))
            self.semgrep_stub = SemgrepStub(grpc.secure_channel(self.url, credentials))
            self.vuln_stub = VulnerabilitiesStub(grpc.secure_channel(self.url, credentials))
            self.provenance_stub = GeoProvenanceStub(grpc.secure_channel(self.url, credentials))
            self.scanning_stub = ScanningStub(grpc.secure_channel(self.url, credentials))

    @classmethod
    def _load_cert(cls, cert_file: str) -> bytes:
        with open(cert_file, 'rb') as f:
            return f.read()

    def deps_echo(self, message: str = 'Hello there!') -> str:
        """
        Send Echo message to the Dependency service
        :param self:
        :param message: Message to send (default: Hello there!)
        :return: echo or None
        """
        request_id = str(uuid.uuid4())
        resp: EchoResponse
        try:
            metadata = self.metadata[:]
            metadata.append(('x-request-id', request_id))  # Set a Request ID
            resp = self.dependencies_stub.Echo(EchoRequest(message=message), metadata=metadata, timeout=3)
        except Exception as e:
            self.print_stderr(
                f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
            )
        else:
            # self.print_stderr(f'resp: {resp} - call: {call}')
            # response_id = ""
            # if not call:
            #     self.print_stderr(f'No call to leverage.')
            # for key, value in call.trailing_metadata():
            #     print('Greeter client received trailing metadata: key=%s value=%s' % (key, value))
            #
            # for key, value in call.trailing_metadata():
            #     if key == 'x-response-id':
            #         response_id = value
            # self.print_stderr(f'Response ID: {response_id}. Metadata: {call.trailing_metadata()}')
            if resp:
                return resp.message
            self.print_stderr(f'ERROR: Problem sending Echo request ({message}) to {self.url}. rqId: {request_id}')
        return None

    def crypto_echo(self, message: str = 'Hello there!') -> str:
        """
        Send Echo message to the Cryptography service
        :param self:
        :param message: Message to send (default: Hello there!)
        :return: echo or None
        """
        request_id = str(uuid.uuid4())
        resp: EchoResponse
        try:
            metadata = self.metadata[:]
            metadata.append(('x-request-id', request_id))  # Set a Request ID
            resp = self.crypto_stub.Echo(EchoRequest(message=message), metadata=metadata, timeout=3)
        except Exception as e:
            self.print_stderr(
                f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
            )
        else:
            if resp:
                return resp.message
            self.print_stderr(f'ERROR: Problem sending Echo request ({message}) to {self.url}. rqId: {request_id}')
        return None

    def get_dependencies(self, dependencies: json, depth: int = 1) -> dict:
        if not dependencies:
            self.print_stderr('ERROR: No dependency data supplied to submit to the API.')
            return None
        resp = self.get_dependencies_json(dependencies, depth)
        if not resp:
            self.print_stderr(f'ERROR: No response for dependency request: {dependencies}')
        return resp

    def get_dependencies_json(self, dependencies: dict, depth: int = 1) -> dict:
        """
        Client function to call the rpc for GetDependencies
        :param dependencies: Message to send to the service
        :param depth: depth of sub-dependencies to search (default: 1)
        :return: Server response or None
        """
        if not dependencies:
            self.print_stderr('ERROR: No message supplied to send to gRPC service.')
            return None

        files_json = dependencies.get('files')

        if files_json is None or len(files_json) == 0:
            self.print_stderr('ERROR: No dependency data supplied to send to gRPC service.')
            return None

        def process_file(file):
            request_id = str(uuid.uuid4())
            try:
                file_request = {'files': [file]}

                request = ParseDict(file_request, DependencyRequest())
                request.depth = depth
                metadata = self.metadata[:]
                metadata.append(('x-request-id', request_id))
                self.print_debug(f'Sending dependency data for decoration (rqId: {request_id})...')
                resp = self.dependencies_stub.GetDependencies(request, metadata=metadata, timeout=self.timeout)

                return MessageToDict(resp, preserving_proto_field_name=True)
            except Exception as e:
                self.print_stderr(
                    f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
                )
                return None

        all_responses = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_CONCURRENT_REQUESTS) as executor:
            future_to_file = {executor.submit(process_file, file): file for file in files_json}

            for future in concurrent.futures.as_completed(future_to_file):
                response = future.result()
                if response:
                    all_responses.append(response)

        SUCCESS_STATUS = 'SUCCESS'

        merged_response = {'files': [], 'status': {'status': SUCCESS_STATUS, 'message': 'Success'}}
        for response in all_responses:
            if response:
                if 'files' in response and len(response['files']) > 0:
                    merged_response['files'].append(response['files'][0])
                # Overwrite the status if the any of the responses was not successful
                if 'status' in response and response['status']['status'] != SUCCESS_STATUS:
                    merged_response['status'] = response['status']
        return merged_response

    def get_vulnerabilities_json(self, purls: dict) -> dict:
        """
        Client function to call the rpc for Vulnerability GetVulnerabilities
        :param purls: Message to send to the service
        :return: Server response or None
        """
        if not purls:
            self.print_stderr('ERROR: No message supplied to send to gRPC service.')
            return None
        request_id = str(uuid.uuid4())
        resp: VulnerabilityResponse
        try:
            request = ParseDict(purls, PurlRequest())  # Parse the JSON/Dict into the purl request object
            metadata = self.metadata[:]
            metadata.append(('x-request-id', request_id))  # Set a Request ID
            self.print_debug(f'Sending vulnerability data for decoration (rqId: {request_id})...')
            resp = self.vuln_stub.GetVulnerabilities(request, metadata=metadata, timeout=self.timeout)
        except Exception as e:
            self.print_stderr(
                f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
            )
        else:
            if resp:
                if not self._check_status_response(resp.status, request_id):
                    return None
                resp_dict = MessageToDict(resp, preserving_proto_field_name=True)  # Convert gRPC response to a dict
                del resp_dict['status']
                return resp_dict
        return None

    def get_semgrep_json(self, purls: dict) -> dict:
        """
        Client function to call the rpc for Semgrep GetIssues
        :param purls: Message to send to the service
        :return: Server response or None
        """
        if not purls:
            self.print_stderr('ERROR: No message supplied to send to gRPC service.')
            return None
        request_id = str(uuid.uuid4())
        resp: SemgrepResponse
        try:
            request = ParseDict(purls, PurlRequest())  # Parse the JSON/Dict into the purl request object
            metadata = self.metadata[:]
            metadata.append(('x-request-id', request_id))  # Set a Request ID
            self.print_debug(f'Sending semgrep data for decoration (rqId: {request_id})...')
            resp = self.semgrep_stub.GetIssues(request, metadata=metadata, timeout=self.timeout)
        except Exception as e:
            self.print_stderr(
                f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
            )
        else:
            if resp:
                if not self._check_status_response(resp.status, request_id):
                    return None
                resp_dict = MessageToDict(resp, preserving_proto_field_name=True)  # Convert gRPC response to a dict
                del resp_dict['status']
                return resp_dict
        return None

    def search_components_json(self, search: dict) -> dict:
        """
        Client function to call the rpc for Components SearchComponents
        :param search: Message to send to the service
        :return: Server response or None
        """
        if not search:
            self.print_stderr('ERROR: No message supplied to send to gRPC service.')
            return None
        request_id = str(uuid.uuid4())
        resp: CompSearchResponse
        try:
            request = ParseDict(search, CompSearchRequest())  # Parse the JSON/Dict into the purl request object
            metadata = self.metadata[:]
            metadata.append(('x-request-id', request_id))  # Set a Request ID
            self.print_debug(f'Sending component search data (rqId: {request_id})...')
            resp = self.comp_search_stub.SearchComponents(request, metadata=metadata, timeout=self.timeout)
        except Exception as e:
            self.print_stderr(
                f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
            )
        else:
            if resp:
                if not self._check_status_response(resp.status, request_id):
                    return None
                resp_dict = MessageToDict(resp, preserving_proto_field_name=True)  # Convert gRPC response to a dict
                del resp_dict['status']
                return resp_dict
        return None

    def get_component_versions_json(self, search: dict) -> dict:
        """
        Client function to call the rpc for Components GetComponentVersions
        :param search: Message to send to the service
        :return: Server response or None
        """
        if not search:
            self.print_stderr('ERROR: No message supplied to send to gRPC service.')
            return None
        request_id = str(uuid.uuid4())
        resp: CompVersionResponse
        try:
            request = ParseDict(search, CompVersionRequest())  # Parse the JSON/Dict into the purl request object
            metadata = self.metadata[:]
            metadata.append(('x-request-id', request_id))  # Set a Request ID
            self.print_debug(f'Sending component version data (rqId: {request_id})...')
            resp = self.comp_search_stub.GetComponentVersions(request, metadata=metadata, timeout=self.timeout)
        except Exception as e:
            self.print_stderr(
                f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
            )
        else:
            if resp:
                if not self._check_status_response(resp.status, request_id):
                    return None
                resp_dict = MessageToDict(resp, preserving_proto_field_name=True)  # Convert gRPC response to a dict
                del resp_dict['status']
                return resp_dict
        return None

    def folder_hash_scan(self, request: Dict) -> Optional[Dict]:
        """
        Client function to call the rpc for Folder Hashing Scan

        Args:
            request (Dict): Folder Hash Request

        Returns:
            Optional[Dict]: Folder Hash Response, or None if the request was not succesfull
        """
        return self._call_rpc(
            self.scanning_stub.FolderHashScan,
            request,
            HFHRequest,
            'Sending folder hash scan data (rqId: {rqId})...',
        )

    def _call_rpc(self, rpc_method, request_input, request_type, debug_msg: Optional[str] = None) -> Optional[Dict]:
        """
        Call a gRPC method and return the response as a dictionary

        Args:
            rpc_method (): The gRPC stub method
            request_input (): Either a dict or a gRPC request object.
            request_type (): The type of the gRPC request object.
            debug_msg (str, optional): Debug message template that can include {rqId} placeholder.

        Returns:
            dict: The parsed gRPC response as a dictionary, or None if something went wrong
        """
        request_id = str(uuid.uuid4())

        if isinstance(request_input, dict):
            request_obj = ParseDict(request_input, request_type())
        else:
            request_obj = request_input

        metadata = self.metadata[:] + [('x-request-id', request_id)]

        self.print_debug(debug_msg.format(rqId=request_id))
        try:
            resp = rpc_method(request_obj, metadata=metadata, timeout=self.timeout)
        except grpc.RpcError as e:
            raise ScanossGrpcError(
                f'{e.__class__.__name__} while sending gRPC message (rqId: {request_id}): {e.details()}'
            )

        if resp and not self._check_status_response(resp.status, request_id):
            return None

        resp_dict = MessageToDict(resp, preserving_proto_field_name=True)
        return resp_dict

    def _check_status_response(self, status_response: StatusResponse, request_id: str = None) -> bool:
        """
        Check the response object to see if the command was successful or not
        :param status_response: Status Response
        :return: True if successful, False otherwise
        """

        if not status_response:
            self.print_stderr(f'Warning: No status response supplied (rqId: {request_id}). Assuming it was ok.')
            return True
        self.print_debug(f'Checking response status (rqId: {request_id}): {status_response}')
        status_code: StatusCode = status_response.status
        if status_code > ScanossGrpcStatusCode.SUCCESS:
            ret_val = False  # default to failed
            msg = 'Unsuccessful'
            if status_code == ScanossGrpcStatusCode.SUCCESS_WITH_WARNINGS:
                msg = 'Succeeded with warnings'
                ret_val = True  # No need to fail as it succeeded with warnings
            elif status_code == ScanossGrpcStatusCode.FAILED_WITH_WARNINGS:
                msg = 'Failed with warnings'
            self.print_stderr(f'{msg} (rqId: {request_id} - status: {status_code}): {status_response.message}')
            return ret_val
        return True

    def _get_proxy_config(self):
        """
        Set the grpc_proxy/http_proxy/https_proxy environment variables if PAC file has been specified
        or if an explicit proxy has been supplied
        :param self:
        """
        if self.grpc_proxy:
            self.print_debug('Setting GRPC (grpc_proxy) proxy...')
            os.environ['grpc_proxy'] = self.grpc_proxy
        elif self.proxy:
            self.print_debug('Setting GRPC (http_proxy/https_proxy) proxies...')
            os.environ['http_proxy'] = self.proxy
            os.environ['https_proxy'] = self.proxy
        elif self.pac:
            self.print_debug(f'Attempting to get GRPC proxy details from PAC for {self.orig_url}...')
            resolver = ProxyResolver(self.pac)
            proxies = resolver.get_proxy_for_requests(self.orig_url)
            if proxies:
                self.print_trace(f'Setting proxies: {proxies}')
            os.environ['http_proxy'] = proxies.get('http') or ''
            os.environ['https_proxy'] = proxies.get('https') or ''

    def get_provenance_json(self, purls: dict) -> dict:
        """
        Client function to call the rpc for GetComponentProvenance
        :param purls: Message to send to the service
        :return: Server response or None
        """
        if not purls:
            self.print_stderr('ERROR: No message supplied to send to gRPC service.')
            return None
        request_id = str(uuid.uuid4())
        resp: ContributorResponse
        try:
            request = ParseDict(purls, PurlRequest())  # Parse the JSON/Dict into the purl request object
            metadata = self.metadata[:]
            metadata.append(('x-request-id', request_id))  # Set a Request ID
            self.print_debug(f'Sending data for provenance decoration (rqId: {request_id})...')
            resp = self.provenance_stub.GetComponentContributors(request, metadata=metadata, timeout=self.timeout)
        except Exception as e:
            self.print_stderr(
                f'ERROR: {e.__class__.__name__} Problem encountered sending gRPC message (rqId: {request_id}): {e}'
            )
        else:
            if resp:
                if not self._check_status_response(resp.status, request_id):
                    return None
                resp_dict = MessageToDict(resp, preserving_proto_field_name=True)  # Convert gRPC response to a dict
                return resp_dict
        return None

    def get_provenance_origin(self, request: Dict) -> Optional[Dict]:
        """
        Client function to call the rpc for GetComponentOrigin

        Args:
            request (Dict): GetComponentOrigin Request

        Returns:
            Optional[Dict]: OriginResponse, or None if the request was not successfull
        """
        return self._call_rpc(
            self.provenance_stub.GetComponentOrigin,
            request,
            PurlRequest,
            'Sending data for provenance origin decoration (rqId: {rqId})...',
        )

    def get_crypto_algorithms_for_purl(self, request: Dict) -> Optional[Dict]:
        """
        Client function to call the rpc for GetAlgorithms for a list of purls

        Args:
            request (Dict): PurlRequest

        Returns:
            Optional[Dict]: AlgorithmResponse, or None if the request was not successfull
        """
        return self._call_rpc(
            self.crypto_stub.GetAlgorithms,
            request,
            PurlRequest,
            'Sending data for cryptographic algorithms decoration (rqId: {rqId})...',
        )

    def get_crypto_algorithms_in_range_for_purl(self, request: Dict) -> Optional[Dict]:
        """
        Client function to call the rpc for GetAlgorithmsInRange for a list of purls

        Args:
            request (Dict): PurlRequest

        Returns:
            Optional[Dict]: AlgorithmsInRangeResponse, or None if the request was not successfull
        """
        return self._call_rpc(
            self.crypto_stub.GetAlgorithmsInRange,
            request,
            PurlRequest,
            'Sending data for cryptographic algorithms in range decoration (rqId: {rqId})...',
        )

    def get_encryption_hints_for_purl(self, request: Dict) -> Optional[Dict]:
        """
        Client function to call the rpc for GetEncryptionHints for a list of purls

        Args:
            request (Dict): PurlRequest

        Returns:
            Optional[Dict]: HintsResponse, or None if the request was not successfull
        """
        return self._call_rpc(
            self.crypto_stub.GetEncryptionHints,
            request,
            PurlRequest,
            'Sending data for encryption hints decoration (rqId: {rqId})...',
        )

    def get_encryption_hints_in_range_for_purl(self, request: Dict) -> Optional[Dict]:
        """
        Client function to call the rpc for GetHintsInRange for a list of purls

        Args:
            request (Dict): PurlRequest

        Returns:
            Optional[Dict]: HintsInRangeResponse, or None if the request was not successfull
        """
        return self._call_rpc(
            self.crypto_stub.GetHintsInRange,
            request,
            PurlRequest,
            'Sending data for encryption hints in range decoration (rqId: {rqId})...',
        )

    def get_versions_in_range_for_purl(self, request: Dict) -> Optional[Dict]:
        """
        Client function to call the rpc for GetVersionsInRange for a list of purls

        Args:
            request (Dict): PurlRequest

        Returns:
            Optional[Dict]: VersionsInRangeResponse, or None if the request was not successfull
        """
        return self._call_rpc(
            self.crypto_stub.GetVersionsInRange,
            request,
            PurlRequest,
            'Sending data for cryptographic versions in range decoration (rqId: {rqId})...',
        )

    def load_generic_headers(self):
        """
        Adds custom headers from req_headers to metadata.

        If x-api-key is present and no URL is configured (directly or via
        environment), sets URL to the premium endpoint (DEFAULT_URL2).
        """
        if self.req_headers:  # Load generic headers
            for key, value in self.req_headers.items():
                if key == 'x-api-key':  # Set premium URL if x-api-key header is set
                    if not self.url and not os.environ.get('SCANOSS_GRPC_URL'):
                        self.url = DEFAULT_URL2  # API key specific and no alternative URL, so use the default premium
                    self.api_key = value
                self.metadata.append((key, value))


#
# End of ScanossGrpc Class
#


@dataclass
class GrpcConfig:
    url: str = DEFAULT_URL
    api_key: Optional[str] = SCANOSS_API_KEY
    debug: Optional[bool] = False
    trace: Optional[bool] = False
    quiet: Optional[bool] = False
    ver_details: Optional[str] = None
    ca_cert: Optional[str] = None
    timeout: Optional[int] = DEFAULT_TIMEOUT
    proxy: Optional[str] = None
    grpc_proxy: Optional[str] = None
    pac: Optional[PACFile] = None
    req_headers: Optional[dict] = None


def create_grpc_config_from_args(args) -> GrpcConfig:
    return GrpcConfig(
        url=getattr(args, 'api2url', DEFAULT_URL),
        api_key=getattr(args, 'key', SCANOSS_API_KEY),
        debug=getattr(args, 'debug', False),
        trace=getattr(args, 'trace', False),
        quiet=getattr(args, 'quiet', False),
        ver_details=getattr(args, 'ver_details', None),
        ca_cert=getattr(args, 'ca_cert', None),
        timeout=getattr(args, 'timeout', DEFAULT_TIMEOUT),
        proxy=getattr(args, 'proxy', None),
        grpc_proxy=getattr(args, 'grpc_proxy', None),
    )
